import unittest
from unittest.mock import Mock, patch

from pydfuutil import dfu
from pydfuutil.dfu_load import DFUFile, do_upload, do_dnload


class TestDFULoader(unittest.TestCase):

    def test_dfu_load_do_upload(self):
        xfer_size = 256
        file = DFUFile(name='test_file', file_p=Mock())  # Provide a mock file object
        total_size = 4096

        dif = Mock()
        dif.upload.return_value = bytes(xfer_size)
        dif.get_status.return_value = total_size

        # Mock file_p.write to return the length of the data written
        with patch.object(file.file_p, 'write', return_value=xfer_size) as mock_write:
            result = do_upload(dif, xfer_size, file, total_size)

        # Assertions
        self.assertEqual(result, total_size)  # Assuming expected_size bytes are received

    def test_dfu_load_do_dnload(self):

        xfer_size = 1024
        result = dfu.StatusRetVal(dfu.Status.OK, 100, dfu.State.DFU_DOWNLOAD_IDLE, 0)

        dif = Mock()
        dif.download.return_value = xfer_size
        dif.get_status.return_value = result

        file = DFUFile(name='test_file', file_p=Mock(), size=xfer_size, suffix_len=0)
        quirks = 0

        # Mock file_p.readinto to return the length of the data read
        with patch.object(file.file_p, 'readinto', return_value=xfer_size) as mock_readinto:
            result = do_dnload(dif, xfer_size, file, quirks)

        # Assertions
        self.assertEqual(result, xfer_size)  # Assuming xfer_size bytes are sent


if __name__ == '__main__':
    unittest.main()



"""
USB Device Firmware Update Implementation for OpenPCD
Protocol definitions for USB DFU
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)

This ought to be compliant to the USB DFU Spec 1.0 as available from
https://www.usb.org/developers/devclass_docs/usbdfu10.pdf

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
from dataclasses import dataclass
from enum import IntEnum

import usb.util


USB_DT_DFU = 0x21
USB_DT_DFU_SIZE = 9
USB_TYPE_DFU = usb.util.CTRL_TYPE_CLASS | usb.util.CTRL_RECIPIENT_INTERFACE


class BmAttributes(IntEnum):
    """Enum of DFU_FUNC_DESCRIPTOR's BmAttributes"""
    USB_DFU_CAN_DOWNLOAD = 0x1
    USB_DFU_CAN_UPLOAD = 0x2
    USB_DFU_MANIFEST_TOL = 0x3
    USB_DFU_WILL_DETACH = 0x4


# pylint: disable=invalid-name
@dataclass
class FuncDescriptor:
    """USB_DFU_FUNC_DESCRIPTOR's'"""
    bLength: int = 0
    bDescriptorType: int = 0
    bmAttributes: BmAttributes = 0
    wDetachTimeOut: int = 0
    wTransferSize: int = 0
    bcdDFUVersion: int = 0



# DFU class-specific requests (Section 3, DFU Rev 1.1)
class UsbReqDfu(IntEnum):
    """Dfu requests"""
    DETACH = 0x00
    DNLOAD = 0x01
    UPLOAD = 0x02
    GETSTATUS = 0x03
    CLRSTATUS = 0x04
    GETSTATE = 0x05
    ABORT = 0x06


# DFU_GETSTATUS bStatus values (Section 6.1.2, DFU Rev 1.1)
class DFUStatus(IntEnum):
    """Dfu statuses"""
    OK = 0x00
    ERROR_TARGET = 0x01
    ERROR_FILE = 0x02
    ERROR_WRITE = 0x03
    ERROR_ERASE = 0x04
    ERROR_CHECK_ERASED = 0x05
    ERROR_PROG = 0x06
    ERROR_VERIFY = 0x07
    ERROR_ADDRESS = 0x08
    ERROR_NOTDONE = 0x09
    ERROR_FIRMWARE = 0x0a
    ERROR_VENDOR = 0x0b
    ERROR_USBR = 0x0c
    ERROR_POR = 0x0d
    ERROR_UNKNOWN = 0x0e
    ERROR_STALLEDPKT = 0x0f


class DFUState(IntEnum):
    """Dfu states"""
    APP_IDLE = 0x00
    APP_DETACH = 0x01
    DFU_IDLE = 0x02
    DFU_DOWNLOAD_SYNC = 0x03
    DFU_DOWNLOAD_BUSY = 0x04
    DFU_DOWNLOAD_IDLE = 0x05
    DFU_MANIFEST_SYNC = 0x06
    DFU_MANIFEST = 0x07
    DFU_MANIFEST_WAIT_RESET = 0x08
    DFU_UPLOAD_IDLE = 0x09
    DFU_ERROR = 0x0a



"""
Helper functions for reading the memory map in a device
following the ST DfuSe 1.1a specification.
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)
"""
import logging
import re
from dataclasses import dataclass, field
from enum import IntFlag
from typing import Iterator

from pydfuutil.logger import logger

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])


class DFUSE(IntFlag):
    """DFUSE read/write flags"""
    READABLE = 0x1
    ERASABLE = 0x2
    WRITEABLE = 0x4


@dataclass
class MemSegment:
    """
    Memory segment

    We're using `segment_stack`
    instead of `segment_list`
    to prevent misunderstanding
    with python's `list` type
    """

    start: int = 0
    end: int = 0
    pagesize: int = 0
    mem_type: int = 0
    next: 'MemSegment' = field(default=None)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'MemSegment':
        """
        Create a MemSegment stack from binary data
        :param data:
        :return MemSegment:
        """
        return cls(
            *data[:4],
            next=MemSegment.from_bytes(data[4:])
            if len(data) > 4 else None
        )

    def __iter__(self) -> Iterator['MemSegment']:
        current = self
        while current is not None:
            yield current
            current = current.next

    def __next__(self):
        if self.next is not None:
            return self.next
        raise StopIteration

    def __len__(self):
        ret = 1
        if self.next is not None:
            ret += self.next.__len__()
        return ret

    def __bytes__(self) -> bytes:
        data = bytes((
            self.start,
            self.end,
            self.pagesize,
            self.mem_type
        ))
        if self.next:
            data += self.next.__bytes__()
        return data

    def append(self, segment: 'MemSegment') -> None:
        """
        Append the other segment to the stack
        :param segment: MemSegment
        """
        new_segment = MemSegment(
            segment.start,
            segment.end,
            segment.pagesize,
            segment.mem_type
        )
        if not self.next:
            self.next = new_segment
        else:
            self.next.append(new_segment)

    def find(self, address: int) -> ['MemSegment', None]:
        """
        Find a memory segment in the stack containing the given element.
        :param address: MemSegment address for in the stack.
        """
        if self.start <= address <= self.end:
            return self
        if self.next is not None:
            return self.next.find(address)
        return None

    # def free(self) -> None:
    #     """Useless cause of garbage collector"""
    #     raise NotImplementedError("Useless cause of garbage collector")


def add_segment(segment_stack: [MemSegment, None], segment: MemSegment) -> MemSegment:
    """
    :param segment_stack:
    :param segment:
    :return: 0 if ok
    """
    new_element = MemSegment(segment.start, segment.end, segment.pagesize, segment.mem_type)

    if not segment_stack:
        # stack can be empty on the first call
        return new_element

    # find the last element in the stack
    segment_stack.append(new_element)
    return segment_stack


def find_segment(segment_stack: [MemSegment, None], address: int) -> [MemSegment, None]:
    """
    Find a memory segment in the stack containing the given element.

    :param segment_stack: List of MemSegment instances.
    :param address: MemSegment address for in the stack.
    :return: MemSegment instance if found, None otherwise.
    """
    if not segment_stack:
        return None
    return segment_stack.find(address)


# def free_segment_list(segment_stack: MemSegment) -> None:
#     """Useless cause of garbage collector"""
#     raise NotImplementedError("Useless cause of garbage collector")


# Parse memory map from interface descriptor string
# encoded as per ST document UM0424 section 4.3.2.

def parse_memory_layout(intf_desc: [str, bytes], verbose: bool = False) -> [MemSegment, None]:
    """
    Parse memory map from interface descriptor string
    encoded as per ST document UM0424 section 4.3.2.
    :param intf_desc:
    :param verbose:
    :return: MemSegment instance
    """

    _logger.setLevel(logging.DEBUG if verbose else logging.INFO)

    if isinstance(intf_desc, bytes):
        intf_desc = intf_desc.decode('ascii')

    count: int = 0
    segment_stack: [MemSegment, None] = None
    address: [int, None] = None

    match = re.match(r'@([^/]+)', intf_desc)
    if match is None:
        _logger.error(f"Could not read name, name={match}")
        return None
    name = match.group(1)
    _logger.info(f"DfuSe interface name: {name}")

    intf_desc = intf_desc[match.end():]

    # while per segment
    while (match := re.match(r'/0x(\d+)/', intf_desc)) is not None:
        address = int(match.group(1), 16)

        intf_desc = intf_desc[match.end():]

        # while per address
        while (match := re.match(r'(\d+)\*(\d+)(\w)(\w)[,/]?', intf_desc)) is not None:
            _sectors, _size, multiplier, type_string = match.groups()
            sectors, size = int(_sectors), int(_size)

            _logger.debug(f"{sectors=}, {size=}, {multiplier=}, {type_string=}")

            intf_desc = intf_desc[match.end():]
            count += 1
            mem_type = ord(type_string)

            if multiplier == 'B':
                pass
            elif multiplier == 'K':
                size *= 1024
            elif multiplier == 'M':
                size *= 1024 * 1024
            elif multiplier in ('a', 'b', 'c', 'd', 'e', 'f', 'g'):
                if not mem_type:
                    _logger.warning(f"Non-valid multiplier {multiplier}, "
                                   "interpreted as type identifier instead")
                    mem_type = multiplier

            # fallthrough if mem_type was already set
            else:
                _logger.warning(f"Non-valid multiplier {multiplier} assuming bytes")

            if not mem_type:
                _logger.warning(f"No valid type for segment {count}")

            segment_stack = add_segment(
                segment_stack,
                MemSegment(
                    address,
                    address + sectors * size - 1,
                    size,
                    mem_type & 7
                )
            )

            _logger.debug(f"Memory segment at "
                         f"0x{address:08x} {sectors} x {size} = {sectors * size} "
                         f"({'r' if mem_type & DFUSE.READABLE else ''}"
                         f"{'e' if mem_type & DFUSE.ERASABLE else ''}"
                         f"{'w' if mem_type & DFUSE.WRITEABLE else ''})")

            address += sectors * size

        _logger.debug(f"Parsed details of {count} segments")

    if address is None:
        _logger.error(f"Could not read address, {address=}")

    return segment_stack



"""
DFU transfer routines
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)

This is supposed to be a general DFU implementation, as specified in the
USB DFU 1.0 and 1.1 specification.

The code was originally intended to interface with a USB device running the
"sam7dfu" firmware (see https://www.openpcd.org/) on an AT91SAM7 processor.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
import logging

import usb

from pydfuutil import dfu
from pydfuutil.dfu_file import DFUFile
from pydfuutil.logger import logger
from pydfuutil.portable import milli_sleep
from pydfuutil.progress import Progress
from pydfuutil.quirks import QUIRK_POLLTIMEOUT, DEFAULT_POLLTIMEOUT

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])


def do_upload(dif: dfu.DfuIf,
              xfer_size: int,
              file: DFUFile = None,
              expected_size: int = -1) -> [int, bytes]:
    """
    Uploads data from DFU device from special page
    :param dif: dfu.dfu_if
    :param xfer_size: chunk size
    :param file: optional - DFUFile object
    :param expected_size: optional - total bytes expected to be uploaded
    :return: uploaded bytes or error code
    """

    _logger.info("Copying data from DFU device to PC")

    total_bytes = 0
    transaction = dfu.TRANSACTION  # start page
    buf = bytearray(xfer_size)

    with Progress() as progress:
        progress_total = expected_size if expected_size >= 0 else None
        progress.start_task(
            description="Starting upload",
            total=progress_total
        )
        # ret = 0  # need there?
        try:
            while True:
                rc = dif.upload(transaction, buf)

                if len(rc) < 0:
                    _logger.error("Error during upload")
                    ret = rc
                    break

                if file:
                    write_rc = file.file_p.write(rc)
                    # TODO: replace to dfu_file_write_crc

                    if write_rc < len(rc):
                        _logger.error(f'Short file write: {write_rc}')
                        ret = total_bytes
                        break

                total_bytes += len(rc)

                if total_bytes < 0:
                    raise IOError("Received too many bytes (wraparound)")


                transaction += 1
                progress.update(advance=len(rc), description="Uploading...")

                # last block, return
                if (len(rc) < xfer_size) or (total_bytes >= expected_size >= 0):
                    ret = total_bytes
                    break

            progress.update(description='Upload finished!')

            _logger.debug(f"Received a total of {total_bytes} bytes")

            if expected_size != 0 and total_bytes != expected_size:
                _logger.warning("Unexpected number of bytes uploaded from device")

            return ret
        except IOError as e:
            _logger.error(e)
            return -1


# pylint: disable=too-many-branches
def do_dnload(dif: dfu.DfuIf, xfer_size: int, file: DFUFile, quirks: int) -> int:
    """
    :param dif: DfuIf instance
    :param xfer_size: transaction size
    :param file: DFUFile instance
    :param quirks: quirks  TODO: replace to DFUFile011 with quirks
    verbose: is verbose useless cause of using python's logging
    :return:
    """

    bytes_sent = 0
    # # TODO: new one
    # buf = file.firmware
    # expected_size = file.size.total - file.size.suffix;
    # bytes_sent = 0
    buf = bytearray(xfer_size)

    _logger.info("Copying data from PC to DFU device")

    try:

        with Progress() as progress:
            total_size = file.size - file.suffix_len  # TODO: replace to expected_size
            progress.start_task(
                description="Starting download",
                total=total_size if total_size >= 0 else None
            )

            while bytes_sent < file.size - file.suffix_len:
                # Note: no idea what's there
                # bytes_left = file.size - file.suffix_len - bytes_sent
                # chunk_size = min(bytes_left, xfer_size)

                if (ret := file.file_p.readinto(buf)) < 0:  # Handle read error
                    raise IOError(f"Error reading file: {file.name}")

                ret = dif.download(ret, buf[:ret] if ret else None)

                if ret < 0:
                    raise IOError("Error during download")
                bytes_sent += ret

                while True:
                    if int(status := dif.get_status()) < 0:
                        raise IOError("Error during download get_status")
                    if status.bState in (dfu.State.DFU_DOWNLOAD_IDLE, dfu.State.DFU_ERROR):
                        break
                    # Wait while the device executes flashing
                    milli_sleep(
                        DEFAULT_POLLTIMEOUT
                        if quirks & QUIRK_POLLTIMEOUT
                        else status.bwPollTimeout
                    )

                if status.bStatus != dfu.Status.OK:
                    logger.error("Transfer failed!")
                    logger.info(f"state({status.bState}) = {status.bState.to_string()}, "
                                f"status({status.bStatus}) = {status.bStatus.to_string()}")
                    raise IOError("Downloading failed!")

                progress.update(description="Downloading...", advance=xfer_size//1000)

            # Send one zero-sized download request to signalize end
            if dif.download(dfu.TRANSACTION, bytes()) < 0:
                raise IOError("Error sending completion packet")

            progress.update(description="Download finished!")
            _logger.info("finished!")
            _logger.debug(f"Sent a total of {bytes_sent} bytes")

            # Transition to MANIFEST_SYNC state
            if int(status := dif.get_status()) < 0:
                raise IOError("Unable to read DFU status")
            _logger.info(f"state({status.bState}) = {status.bState.to_string()}, "
                         f"status({status.bStatus}) = {status.bStatus.to_string()}")

            if not quirks & QUIRK_POLLTIMEOUT:
                milli_sleep(status.bwPollTimeout)

            # Deal correctly with ManifestationTolerant=0 / WillDetach bits
            while status.bState in (dfu.State.DFU_MANIFEST_SYNC, dfu.State.DFU_MANIFEST):
                # Some devices need some time before we can obtain the status
                milli_sleep(1000)
                if int(status := dif.get_status()) < 0:
                    raise IOError("Unable to read DFU status")
                _logger.info(f"state({status.bState}) = {status.bState.to_string()}, "
                             f"status({status.bStatus}) = {status.bStatus.to_string()}")

            if status.bState == dfu.State.DFU_IDLE:
                _logger.info("Done!")
        return bytes_sent

    except IOError as err:
        _logger.error(err)
        return -1


def init() -> None:
    """Init dfu_load props"""
    dfu.debug(dfu.DEBUG)
    dfu.init(dfu.TIMEOUT)



import os.path
import unittest

from pydfuutil.dfu_file import *
from pydfuutil.dfu_file import crc32_byte


class TestDFUFile(unittest.TestCase):

    """
    unsigned char _suffix[] = {
         0x00, /* bcdDevice lo */
         0x00, /* bcdDevice hi */
         0x00, /* idProduct lo */
         0x00, /* idProduct hi */
         0x00, /* idVendor lo */
         0x00, /* idVendor hi */
         0x00, /* bcdDFU lo */
         0x01, /* bcdDFU hi */
         'U', /* ucDfuSignature lsb */
         'F', /* ucDfuSignature --- */
         'D', /* ucDfuSignature msb */
         16, /* bLength for this version */
         0x00, /* dwCRC lsb */
         0x00, /* dwCRC --- */
         0x00, /* dwCRC --- */
         0x00 /* dwCRC msb */
        };
    """

    def setUp(self):
        # You may need to adjust these paths based on your actual implementation
        self.sample_file_path = os.path.join(os.path.dirname(__file__), "sample_file.bin")
        self.output_file_path = os.path.join(os.path.dirname(__file__), "output_file.bin")

    def test_crc32_byte(self):
        # You may need to adjust the expected result based on your specific CRC calculation
        result = crc32_byte(0, 0)
        self.assertEqual(result, 0x0)

    def test_parse_dfu_suffix(self):
        dfu_file = DFUFile(self.sample_file_path)
        result = parse_dfu_suffix(dfu_file)
        # Adjust the expected result based on the expected behavior of parse_dfu_suffix
        self.assertTrue(result >= 0)

    def test_generate_dfu_suffix(self):
        dfu_file = DFUFile(self.output_file_path)
        dfu_file.bcdDevice = 0x0100
        dfu_file.idProduct = 0x1234
        dfu_file.idVendor = 0x5678

        # Create a longer sample file
        with open(self.output_file_path, 'wb') as file:
            file.write(b'\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A' + b'\x00' * 100)

        # # Assuming that parse_dfu_suffix works correctly, we parse the sample file first
        parse_result = parse_dfu_suffix(dfu_file)
        self.assertTrue(parse_result >= 0)

        # Then, we generate a new DFU suffix for the output file
        generate_result = generate_dfu_suffix(dfu_file)
        # Adjust the expected result based on the expected behavior of generate_dfu_suffix
        self.assertTrue(generate_result >= 0)

        parse_result = parse_dfu_suffix(dfu_file)
        self.assertTrue(parse_result >= 0)


if __name__ == '__main__':
    unittest.main()



"""
This implements the TI Stellaris DFU
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)

as per the Application Update Using the USB Device Firmware Upgrade Class
(Document AN012373)
https://www.ti.com/general/docs/lit/getliterature.tsp?literatureNumber=spma003&fileType=pdf
"""

from pydfuutil.dfu_file import DFUFile, parse_dfu_suffix, generate_dfu_suffix
from pydfuutil.logger import logger

__all__ = ('dfu_prefix',
           'add_prefix',
           'remove_prefix',
           'check_prefix',
           'parse_dfu_suffix',
           'generate_dfu_suffix')

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])

# dfu_prefix payload length excludes prefix and suffix


dfu_prefix = (
    0x01,  # STELLARIS_DFU_PROG
    0x00,  # Reserved
    0x00,  # LSB start address / 1024
    0x20,  # MSB start address / 1024
    0x00,  # LSB file payload length
    0x00,  # Byte 2 file payload length
    0x00,  # Byte 3 file payload length
    0x00,  # MSB file payload length
)


def add_prefix(file: DFUFile, address: int) -> int:
    """
    Add TI Stellaris DFU prefix to a binary file.

    :param file: DFUFile instance representing the file to be modified.
    :param address: Integer representing the starting address for the TI Stellaris DFU prefix.
    :return: 0 on success, -1 on error.
    """

    try:
        # Get file length
        file.file_p.seek(0, 2)
        length = file.file_p.tell()
        file.file_p.seek(0, 0)

        # Read file content
        data = file.file_p.read()

        # Allocate buffer
        lmdfu_dfu_prefix_buf = bytearray([0] * 16)

        # Fill Stellaris dfu_prefix with correct data
        addr = address // 1024
        lmdfu_dfu_prefix_buf[2] = addr & 0xff
        lmdfu_dfu_prefix_buf[3] = (addr >> 8) & 0xff
        lmdfu_dfu_prefix_buf[4] = length & 0xff
        lmdfu_dfu_prefix_buf[5] = (length >> 8) & 0xff
        lmdfu_dfu_prefix_buf[6] = (length >> 16) & 0xff
        lmdfu_dfu_prefix_buf[7] = (length >> 24) & 0xff

        # Write TI Stellaris DFU prefix to the file
        file.file_p.seek(0)
        file.file_p.write(lmdfu_dfu_prefix_buf)

        # Write file content after the TI Stellaris DFU prefix
        file.file_p.write(data)

        _logger.info("TI Stellaris DFU prefix added.")

        return 0
    except Exception as e:
        _logger.error(f"Error: {e}")
        return -1


def remove_prefix(file: DFUFile) -> int:
    """
    Remove TI Stellaris DFU prefix from a binary file.

    :param file: DFUFile instance representing the file to be modified.
    :return: 0 on success, -1 on error.
    """

    try:
        _logger.info("Remove TI Stellaris prefix")

        # Get file length
        file.file_p.seek(0, 2)
        length = file.file_p.tell()
        file.file_p.seek(0, 0)

        # Read file content
        data = file.file_p.read()

        # Check if the file has enough data to contain the prefix
        if length < 16:
            _logger.error("Error: File does not contain a valid prefix.")
            return -1

        # Truncate the file
        file.file_p.truncate(0)
        file.file_p.seek(0)

        # Write data without the TI Stellaris prefix
        file.file_p.write(data[16:])

        _logger.info("TI Stellaris prefix removed")
        return 0

    except Exception as e:
        _logger.error(f"Error: {e}")
        return -1


def check_prefix(file: DFUFile) -> int:
    """
    Check if a binary file contains a valid TI Stellaris DFU prefix.

    :param file: DFUFile instance representing the file to be checked.
    :return: 0 if not a valid prefix, 1 if a valid prefix, -1 on error.
    """

    try:
        _logger.info("Check TI Stellaris prefix")

        # Allocate buffer for reading the prefix
        data = bytearray(16)

        # Read prefix from the file
        ret = file.file_p.readinto(data)
        if ret < 16:
            _logger.error("Error: Could not read prefix")
            return -1

        # Check if it's a valid TI Stellaris DFU prefix
        if data[0] != 0x01 or data[1] != 0x00:
            _logger.info("Not a valid TI Stellaris DFU prefix")
            ret = 0
        else:
            _logger.info("Possible TI Stellaris DFU prefix with the following properties:")
            address = 1024 * (data[3] << 8 | data[2])
            payload_length = data[4] | data[5] << 8 | data[6] << 16 | data[7] << 24
            _logger.info(f"Address:        0x{address:08X}")
            _logger.info(f"Payload length: {payload_length}")

        # Rewind the file
        file.file_p.seek(0)

        return ret

    except Exception as e:
        _logger.error(f"Error: {e}")
        return -1



"""Progress bar utilities"""


import sys
from abc import ABC, abstractmethod



try:
    from rich import progress as RICH_PROGRESS
except ImportError:
    try:
        from pip._vendor.rich import progress as RICH_PROGRESS
    except ImportError:
        RICH_PROGRESS = None

try:
    from tqdm import tqdm as TQDM_PROGRESS
except ImportError:
    TQDM_PROGRESS = None


import logging


_logger = logging.getLogger('progress')
_logger.setLevel(logging.INFO)


class AbstractProgressBackend(ABC):
    """Abstract class for progress bar backends"""

    def __init__(self):
        ...

    @abstractmethod
    def start(self):
        """
        Progress backend initialisation
        """

    @abstractmethod
    def start_task(self, *, description: str = None, total: int = None):
        """
        Start progress task
        :param description:
        :param total:
        """

    @abstractmethod
    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        """
        Update progress task
        :param description:
        :param advance:
        :param completed:
        """

    @abstractmethod
    def fail(self):
        """Runs on Progress.__exit__ if ctx raises exception"""

    @abstractmethod
    def stop(self):
        """Stop progressbar backend"""


class NoProgressBarBackend(AbstractProgressBackend):
    """progress bar backend that does nothing"""

    def start(self):
        pass

    def start_task(self, *, description: str = None, total: int = None):
        pass

    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        pass

    def fail(self):
        pass

    def stop(self):
        pass


class AsciiBackend(AbstractProgressBackend):
    """ASCII based progress bar backend"""

    BAR_WIDTH = 30
    REDIRECT_STD = True

    def __init__(self):
        super().__init__()
        self._value = None
        self._total = None
        self._rate = None
        self._fail = None

    def _print(self, symbol: str):
        if self.REDIRECT_STD:
            sys.stdout.write(symbol)
            sys.stdout.flush()
        else:
            print(symbol, end="", flush=True)

    def start(self):
        pass

    def start_task(self, *, description: str = None, total: int = None):
        self._value = 0
        self._total = total
        self._fail = False
        if self._total:
            self._rate = self._total / self.BAR_WIDTH
        self._print(f"{description} [")

    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        if self._total:

            if advance:
                self._value += advance
                if self._rate != 0 and advance % self._rate != 0:
                    self._print("━" * int(advance / self._rate))
            if completed:
                if completed < self._value:
                    raise ValueError(f"The progress can't run backward! "
                                     f"{completed} < {self._value}")
                prev_value = self._value
                self._value = completed
                if (delta := self._value - prev_value) > 0:
                    if delta % self._rate != 0:
                        self._print("━" * int(delta / self._rate))
            if self._total <= self._value and (advance or completed):
                self._print("] Complete!\n")
        else:
            self._print("━")

    def fail(self):
        self._print("] Failed!\n")

    def stop(self):
        if not self._total and not self._fail:
            self._print("] Complete!\n")


class TqdmBackend(AbstractProgressBackend):
    """tqdm based progress bar backend"""

    BAR_FORMAT = ("{desc} {bar:20} {percentage:3.0f}% "
                  "{remaining} {n_fmt}/{total_fmt} bytes {rate_fmt}")
    BAR_FORMAT_INF = "{desc} {postfix} {n_fmt}/{total_fmt} bytes {rate_fmt}"

    def __init__(self):
        super().__init__()
        self._progress = None
        self._value = None
        self._spinner_state = None

    def start(self):
        pass

    def start_task(self, *, description: str = None, total: int = None):
        self._value = 0
        if total:
            self._progress = TQDM_PROGRESS(
                total=total,
                unit=' bytes',
                desc=description,
                bar_format=TqdmBackend.BAR_FORMAT,
                ascii=' ━',
            )
        else:
            self._progress = TQDM_PROGRESS(
                total=float('inf'),
                unit=' bytes',
                desc=description,
                bar_format=TqdmBackend.BAR_FORMAT_INF,
                postfix=""
            )
            self._spinner_state = 0

    def _spin(self, complete=False):
        if complete:
            self._progress.postfix = "━" * 20
            return

        symbol = "=▶" if self._spinner_state % 2 == 1 else "▶="
        self._progress.postfix = symbol * 10
        self._spinner_state += 1

    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        if description:
            self._progress.desc = description

        if advance:
            self._progress.update(advance)
        if completed:
            self._progress.n = completed

        if self._progress.total:
            if self._progress.total == self._value and (advance or completed):
                self._progress.desc = description
        else:
            self._spin()
        self._progress.refresh()

    def fail(self):
        # self._progress.colour = "red"
        # self._progress.refresh()
        pass

    def stop(self):
        if not self._progress.total:
            self._spin(complete=True)
        self._progress = None


class RichBackend(AbstractProgressBackend):
    """Rich.progress based progress bar backend"""

    def __init__(self):
        super().__init__()
        self._progress: [RICH_PROGRESS.Progress, None] = None
        self._task_id = None
        self._fail = None

    def start(self):
        pass

    def _prepare(self):
        self._progress = RICH_PROGRESS.Progress(
            RICH_PROGRESS.TextColumn(
                "[progress.description]{task.description}"),
            RICH_PROGRESS.BarColumn(20),
            RICH_PROGRESS.TaskProgressColumn(),
            RICH_PROGRESS.TimeRemainingColumn(),
            RICH_PROGRESS.DownloadColumn(),
            RICH_PROGRESS.TransferSpeedColumn(),
        )
        self._progress.start()

    def start_task(self, *, description: str = None, total: int = None):
        self._prepare()
        self._fail = False
        kwargs = {}
        if description is not None:
            kwargs["description"] = f"[#F92672]{description}"
        kwargs["total"] = total
        self._task_id = self._progress.add_task(start=True, **kwargs)

    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        kwargs = {}
        if description is not None:
            # kwargs["description"] = f"[rgb(249,38,114)]{description}"
            kwargs["description"] = f"[#F92672]{description}"
        if advance is not None:
            kwargs["advance"] = advance
        if completed is not None:
            kwargs["completed"] = completed
        self._progress.update(self._task_id, **kwargs)
        t = self._progress.tasks[self._task_id]
        if t.total is not None and t.completed >= t.total:
            desc = t.description.split(']')[-1]
            self._progress.update(self._task_id,
                                  description=f"[#729C1F]{desc}")

    def fail(self):
        self._fail = True
        t = self._progress.tasks[self._task_id]
        desc = t.description.split(']')[-1]
        self._progress.update(self._task_id,
                              description=f"[red]{desc}")

    def stop(self):
        if not self._fail:
            t = self._progress.tasks[self._task_id]
            desc = t.description.split(']')[-1]
            self._progress.update(self._task_id,
                                  description=f"[#729C1F]{desc}",
                                  total=t.completed)
        # # uncomment to hide bar after complete
        # self._progress.remove_task(self._task_id)
        self._progress.stop()
        self._progress = None


def find_backend():
    """searching for an available backend"""
    if RICH_PROGRESS is not None:
        return RichBackend
    if TQDM_PROGRESS is not None:
        return TqdmBackend
    return AsciiBackend


class Progress:
    """
    High leveled progress bar class
    Use this as a context
    """

    __DEFAULT_BACKEND = find_backend()

    def __init__(self, backend=None):
        if backend is None:
            backend = Progress.__DEFAULT_BACKEND
        self._backend = backend()

    @classmethod
    def set_default_backend(cls, backend: type[AbstractProgressBackend]):
        """
        Sets the default progressbar backend
        :param backend:
        :return:
        """
        if not issubclass(backend, AbstractProgressBackend):
            raise TypeError("Invalid backend")
        cls.__DEFAULT_BACKEND = backend

    def start(self):
        """
        Progress backend initialisation
        """
        self._backend.start()

    def start_task(self, *, description: str = None, total: int = None):
        """
        Start progress task
        :param description:
        :param total:
        """
        self._backend.start_task(description=description, total=total)

    def update(self, *, description: str = None, advance: int = None, completed: int = None):
        """
        Update progress task
        :param description:
        :param advance:
        :param completed:
        """
        self._backend.update(description=description, advance=advance, completed=completed)

    def stop(self):
        """Stop progressbar backend"""
        self._backend.stop()

    def fail(self):
        """
        Runs on progress __exit__ if ctx raises exception
        """
        self._backend.fail()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self.fail()
            raise exc_type(exc_value)
        self.stop()
        return True


__all__ = (
    'Progress',
    'AsciiBackend',
    'RichBackend',
    'TqdmBackend',
    'RICH_PROGRESS',
    'TQDM_PROGRESS',
    'AbstractProgressBackend',
    'NoProgressBarBackend',
    'find_backend'
)



import unittest
from pydfuutil.dfuse_mem import MemSegment, add_segment, parse_memory_layout


class TestDfuSeMem(unittest.TestCase):

    def test_parse_build(self):
        mem_seq = MemSegment.from_bytes(bytes(12))
        buf = bytes(mem_seq)
        self.assertEqual(buf, bytes(12))

    def test_append(self):
        mem_seq = MemSegment.from_bytes(bytes(12))
        seg = MemSegment.from_bytes(bytes(4))
        mem_seq.append(seg)
        buf = bytes(mem_seq)
        self.assertEqual(buf, bytes(16))
        self.assertEqual(len(mem_seq), 4)

    def test_iter(self):
        mem_seq = MemSegment.from_bytes(bytes(12))
        self.assertEqual(len(tuple(mem_seq)), len(mem_seq))

    def test_find(self):
        mem_seq = MemSegment.from_bytes(bytes([1, 2, 0, 0, 10, 20, 0, 0, 100, 200, 0, 0]))
        self.assertEqual(mem_seq.find(10),
                         mem_seq.from_bytes(bytes([10, 20, 0, 0, 100, 200, 0, 0])))
        self.assertEqual(mem_seq.find(150),
                         mem_seq.from_bytes(bytes([100, 200, 0, 0])))
        self.assertEqual(mem_seq.find(300), None)

    def test_free(self):
        mem_seq = MemSegment.from_bytes(bytes(10))
        del mem_seq
        self.assertNotIn('mem_seq', locals())

    def test_add_segment_empty_list(self):
        # Test adding a segment to an empty list
        seqment_sequence = None
        segment = MemSegment(0, 10, 4, 0)
        seqment_sequence = add_segment(seqment_sequence, segment)
        self.assertNotEqual(seqment_sequence, None)
        self.assertEqual(seqment_sequence.end, 10)
        self.assertEqual(seqment_sequence.next, None)

    def test_add_segment_nonempty_list(self):
        # Test adding a segment to a non-empty list
        seqment_sequence = MemSegment(0, 10, 4, 0)
        segment = MemSegment(12, 20, 4, 0)
        seqment_sequence = add_segment(seqment_sequence, segment)

        self.assertNotEqual(seqment_sequence, None)
        self.assertNotEqual(seqment_sequence.next, None)
        self.assertEqual(seqment_sequence.next.start, 12)
        self.assertEqual(seqment_sequence.next.end, 20)

    def test_add_segment_with_existing_next(self):
        # Test adding a segment when the last element in the list already has a next element
        seqment_sequence = MemSegment(0, 10, 4, 0)
        seqment_sequence.next = MemSegment(22, 30, 4, 0)
        segment = MemSegment(12, 20, 4, 0)
        seqment_sequence = add_segment(seqment_sequence, segment)
        self.assertNotEqual(seqment_sequence.next, None)
        self.assertNotEqual(seqment_sequence.next.next, None)
        self.assertEqual(seqment_sequence.next.end, 30)
        self.assertEqual(seqment_sequence.next.next.end, 20)

    def test_parse_memory_layout(self):
        with self.subTest("Test case 1: Valid input"):
            intf_desc = "@Internal Flash/0x08000000/04*016Kg,01*064Kg,01*128Kg"
            result = parse_memory_layout(intf_desc)
            self.assertIsNotNone(result)
            self.assertIsInstance(result, MemSegment)
            self.assertEqual(len(result), 3)

        with self.subTest("Test case 2: Empty input"):
            result = parse_memory_layout("")
            self.assertIsNone(result)

        with self.subTest("Test case 3: Invalid input with missing name"):
            intf_desc = "/0x1000/1*1024B/04*016Kg"
            result = parse_memory_layout(intf_desc)
            self.assertIsNone(result)

        with self.subTest("Test case 4: Invalid input with missing address"):
            intf_desc = "@Internal Flash/1*1024B/"
            result = parse_memory_layout(intf_desc)
            self.assertIsNone(result)

        with self.subTest("Test case 5: Invalid input with missing segment details"):
            intf_desc = "@Internal Flash/0x08000000/"
            result = parse_memory_layout(intf_desc)
            self.assertIsNone(result)  # Should return None, as it's invalid


if __name__ == '__main__':
    unittest.main()


import unittest
from time import sleep

from pydfuutil.progress import *

unittest.TestLoader.sortTestMethodsUsing = None


class TestDfuProgress(unittest.TestCase):

    def _loop(self, backend=None, total=None):
        i = 10

        with Progress(backend) as prog:
            name = (backend.__name__
                    if backend
                    else f"Any")

            prog.start_task(description=f"{name}",
                            total=total)
            while i >= 1:
                sleep(0.1)
                prog.update(advance=1)
                i -= 1
            prog.update(description=f"{name} OK")

    @unittest.skipIf(not TQDM_PROGRESS,
                     "package not installed ImportError/UnboundLocalError/AttributeError")
    def test_tqdm(self):
        self._loop(TqdmBackend, 10)

        with self.subTest("indeterminate"):
            self._loop(TqdmBackend, None)

    @unittest.skipIf(not RICH_PROGRESS,
                     "package not installed ImportError/UnboundLocalError/AttributeError")
    def test_rich(self):
        self._loop(RichBackend, 10)

        with self.subTest("indeterminate"):
            self._loop(RichBackend, None)

    def test_no_progress(self):
        self._loop(NoProgressBarBackend)

    def test_ascii(self):
        with self.subTest("by completed"):
            i, j = 20, 1
            n = AsciiBackend.__name__
            with Progress(AsciiBackend) as prog:
                prog.start_task(description=n,
                                total=i)
                while i >= 0:
                    sleep(0.1)
                    prog.update(completed=21-i)
                    i -= 1 + j
                    j += 1
                prog.update(description=f"{n} OK")

        with self.subTest("by advance"):
            self._loop(AsciiBackend, 10)

        with self.subTest("indeterminate"):
            self._loop(AsciiBackend, None)

    @unittest.skip("rich.errors.LiveError: Only one live display may be active at once")
    def test_autodetect(self):
        self._loop(None)

    def test_exception(self):
        i = 10
        with self.assertRaises(Exception):
            with Progress(AbstractProgressBackend) as prog:
                prog.start_task(description=f"OnError",
                                total=i)
                while i >= 1:
                    sleep(0.1)
                    prog.update(advance=1)
                    i -= 1
                    if i == 5:
                        raise Exception("Something went wrong")


if __name__ == "__main__":
    unittest.main()



import logging
import unittest
from unittest.mock import MagicMock, patch
from pydfuutil.lmdfu import DFUFile, add_prefix, remove_prefix, check_prefix
from pydfuutil.logger import logger

logger.setLevel(logging.DEBUG)


class TestLmdfu(unittest.TestCase):
    @patch('pydfuutil.lmdfu.open')
    def test_add_prefix_success(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.tell.return_value = 100
        mock_open.return_value.__enter__.return_value = file_mock

        address = 0x1000
        result = add_prefix(file_mock, address)

        file_mock.file_p.write.assert_called()
        self.assertEqual(result, 0)

    @patch('pydfuutil.lmdfu.open')
    def test_add_prefix_error(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.tell.side_effect = Exception("File error")
        mock_open.return_value.__enter__.return_value = file_mock

        address = 0x1000
        result = add_prefix(file_mock, address)

        file_mock.file_p.write.assert_not_called()
        self.assertEqual(result, -1)

    @patch('pydfuutil.lmdfu.open')
    def test_remove_prefix_success(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.tell.return_value = 20
        mock_open.return_value.__enter__.return_value = file_mock

        # Mock file content
        file_content = b'TI Stellaris DFU prefix data' + b'file content'
        file_mock.file_p.read.return_value = file_content

        result = remove_prefix(file_mock)

        file_mock.file_p.truncate.assert_called_with(0)
        file_mock.file_p.write.assert_called_with(file_content[16:])
        self.assertEqual(result, 0)

    @patch('pydfuutil.lmdfu.open')
    def test_remove_prefix_error_invalid_prefix(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.tell.return_value = 10
        mock_open.return_value.__enter__.return_value = file_mock

        result = remove_prefix(file_mock)

        file_mock.file_p.truncate.assert_not_called()
        file_mock.file_p.write.assert_not_called()
        self.assertEqual(result, -1)

    @patch('pydfuutil.lmdfu.open')
    def test_remove_prefix_error_exception(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.tell.side_effect = Exception("File error")
        mock_open.return_value.__enter__.return_value = file_mock

        result = remove_prefix(file_mock)

        file_mock.file_p.truncate.assert_not_called()
        file_mock.file_p.write.assert_not_called()
        self.assertEqual(result, -1)

    @patch('pydfuutil.lmdfu.open')
    def test_check_prefix_valid(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.readinto.return_value = 16
        mock_open.return_value.__enter__.return_value = file_mock

        result = check_prefix(file_mock)

        self.assertEqual(result, 0)

    @patch('pydfuutil.lmdfu.open')
    def test_check_prefix_invalid(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.readinto.return_value = 16
        mock_open.return_value.__enter__.return_value = file_mock

        result = check_prefix(file_mock)

        self.assertEqual(result, 0)

    @patch('pydfuutil.lmdfu.open')
    def test_check_prefix_error_read(self, mock_open):
        # Mock file
        file_mock = DFUFile("")
        file_mock.file_p = MagicMock()
        file_mock.file_p.readinto.side_effect = Exception("File error")
        mock_open.return_value.__enter__.return_value = file_mock

        result = check_prefix(file_mock)

        self.assertEqual(result, -1)

if __name__ == '__main__':
    unittest.main()


"""
This implements the ST Microsystems DFU extensions (DfuSe)
as per the DfuSe 1.1a specification (Document UM0391)
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)
"""
import argparse
import errno
import sys
from enum import Enum

import usb.util

from pydfuutil import dfu
from pydfuutil.dfu_file import DFUFile
from pydfuutil.dfuse_mem import find_segment, DFUSE, parse_memory_layout, MemSegment
from pydfuutil.exceptions import GeneralError
from pydfuutil.logger import logger
from pydfuutil.portable import milli_sleep
from pydfuutil.progress import Progress

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])

VERBOSE = False
MEM_LAYOUT: [MemSegment, None] = None

TIMEOUT = 5000


class Command(Enum):
    """DFUSE commands"""
    SET_ADDRESS = 0x1
    ERASE_PAGE = 0x2
    MASS_ERASE = 0x3
    READ_UNPROTECT = 0x4


def quad2uint(p: bytes) -> int:
    """
    Convert a 4-byte sequence into an unsigned integer (little-endian).
    :param p: 4-byte sequence
    :return: Converted unsigned integer
    """
    return int.from_bytes(p, byteorder='little', signed=False)


def parse_options(options: str) -> argparse.Namespace:
    """
    Parse DFU options string and set corresponding flags and values.
    :param options: DFU options string containing address, modifiers, and values.
    :return: None
    """

    parser = argparse.ArgumentParser(description='Parse DFU options')
    parser.add_argument('--address', type=int, help='DFU address')
    parser.add_argument('--force', action='store_true', help='Force option')
    parser.add_argument('--leave', action='store_true', help='Leave option')
    parser.add_argument('--unprotect', action='store_true', help='Unprotect option')
    parser.add_argument('--mass-erase', action='store_true', help='Mass erase option')
    parser.add_argument('--length', type=int, help='Upload length')

    args, _ = parser.parse_known_args(options.split())
    return args


def special_command(dif: dfu.DfuIf, address: int, command: Command) -> int:
    """
    Perform DfuSe-specific commands.
    :param dif: DFU interface
    :param address: Address for the command
    :param command: DfuSe command to execute
    :return: None
    """
    buf = bytearray(5)

    ret: int
    dst: [dict, None]

    try:

        if command == Command.ERASE_PAGE:
            segment = find_segment(MEM_LAYOUT, address)

            if not segment or not segment.mem_type & DFUSE.ERASABLE:
                raise IOError(f"Page at 0x{address:08x} cannot be erased")

            page_size = segment.pagesize
            if VERBOSE > 1:
                _logger.info(
                    f"Erasing page size {page_size} at address 0x{address:08x}, "
                    f"page starting at 0x{address & ~(page_size - 1):08x}")
            buf[0], length = 0x41, 5  # Note: Unused variable 'length'
            # last_erased = address  # Note: useless?
        elif command == Command.SET_ADDRESS:
            if VERBOSE > 2:
                _logger.debug(f"Setting address pointer to 0x{address:08x}")
            buf[0], length = 0x21, 5  # Set Address Pointer command
        elif command == Command.MASS_ERASE:
            buf[0], length = 0x41, 1  # Mass erase command when length = 1
        elif command == Command.READ_UNPROTECT:
            buf[0], length = 0x92, 1
        else:
            raise ValueError(f"Non-supported special command {command}")

        # # overhead
        # buf[1] = address & 0xff; buf[2] = (address >> 8) & 0xff;
        # buf[3] = (address >> 16) & 0xff; buf[4] = (address >> 24) & 0xff

        # for i in range(1, 5):
        #     buf[i] = (address >> (8 * (i - 1))) & 0xFF

        for i in range(0, 4):
            buf[i + 1] = (address >> (8 * i)) & 0xFF

        if download(dif, buf, 0) < 0:
            raise IOError("Error during special command download")

        ret = int(dst := dif.get_status())
        if ret < 0:
            raise IOError("Error during special command get_status")

        if dst.bState != dfu.State.DFU_DOWNLOAD_BUSY:
            raise IOError("Wrong state after command download")

        # Wait while command is executed
        if VERBOSE:
            _logger.info(f"Poll timeout {dst.bwPollTimeout} ms")

        milli_sleep(dst.bwPollTimeout)

        if command == Command.READ_UNPROTECT:
            return ret

        ret = int(dst := dif.get_status())
        if ret < 0:
            _logger.error(
                f"state({dst.bState}) = {dst.bState.to_string()}, "
                f"status({dst.bStatus}) = {dst.bStatus.to_string()}")
            raise IOError("Error during second get_status")

        if dst.bStatus != dfu.Status.OK:
            raise IOError("Command not correctly executed")

        milli_sleep(dst.bwPollTimeout)

        if dif.abort() < 0:
            raise IOError("Error sending dfu abort request")

        ret = int(dst := dif.get_status())
        if ret < 0:
            raise IOError("Error during abort get_status")

        if dst.bState != dfu.State.DFU_IDLE:
            raise IOError("Failed to enter idle state on abort")

    except (ValueError, IOError) as err:
        _logger.error(err)
        raise GeneralError

    milli_sleep(dst.bwPollTimeout)
    return ret


def upload(dif: dfu.DfuIf, data: bytes, transaction: int) -> int:
    """
    UPLOAD request for DfuSe 1.1a

    :param dif: The USB device handle
    :param data: The data buffer to store the uploaded data
    :param transaction: The transaction ID for the upload
    :return: The status of the control transfer
    """
    status = dif.dev.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_IN |
                      usb.util.CTRL_TYPE_CLASS |
                      usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=dfu.Command.UPLOAD,
        wValue=transaction,
        wIndex=dif.interface,
        data_or_wLength=data,
        timeout=TIMEOUT
    )

    if status < 0:
        _logger.error(f"{upload.__name__}: libusb_control_msg returned {status}")

    return status


def download(dif: dfu.DfuIf, data: bytes, transaction: int) -> int:
    """
    DNLOAD request for DfuSe 1.1a

    :param dif: The DFU interface object.
    :param data: The data buffer to download.
    :param transaction: The transaction ID for the download.
    :return: The status of the control transfer.
    """
    status = dif.dev.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_OUT |
                      usb.util.CTRL_TYPE_CLASS |
                      usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=dfu.Command.DNLOAD,
        wValue=transaction,
        wIndex=dif.interface,
        data_or_wLength=data,
        timeout=TIMEOUT
    )

    if status < 0:
        _logger.error(f"{download.__name__}: "
                     f"libusb_control_transfer returned {status}")

    return status


def do_upload(dif: dfu.DfuIf, xfer_size: int, file: DFUFile, dfuse_options: [str, bytes]) -> int:
    """
    :param dif:
    :param xfer_size:
    :param file:
    :param dfuse_options:
    :return:
    """

    # pylint: disable=global-statement
    global MEM_LAYOUT

    total_bytes = 0
    upload_limit = 0
    buf = bytearray(xfer_size)
    transaction = 2

    try:
        if dfuse_options:
            parsed_args = parse_options(dfuse_options)
            if parsed_args.length:
                upload_limit = parsed_args.length
        else:
            raise ValueError("No options provided")

        if parsed_args.address:
            MEM_LAYOUT = parse_memory_layout(dif.alt_name)  # HOTFIX
            if not MEM_LAYOUT:
                raise IOError("Failed to parse memory layout")

            segment = find_segment(MEM_LAYOUT, parsed_args.address)
            if not parsed_args.force and (not segment or not segment.mem_type & DFUSE.READABLE):
                raise IOError(f"Page at 0x{parsed_args.address:08x} is not readable")

            if not upload_limit:
                upload_limit = segment.end - parsed_args.address + 1
                _logger.info(f"Limiting upload to end of memory segment, {upload_limit} bytes")
            special_command(dif, parsed_args.address, Command.SET_ADDRESS)
        else:
            # Bootloader decides the start address, unknown to us
            # Use a short length to lower risk of running out of bounds
            if not upload_limit:
                upload_limit = 0x4000
            _logger.info("Limiting default upload to %i bytes", upload_limit)

    except (ValueError, IOError) as err:
        _logger.error(err)
        return -1

    try:
        with Progress() as progress:
            progress.start_task(
                description="Starting upload",
                total=total_bytes
            )
            while True:
                xfer_size = min(xfer_size, upload_limit - total_bytes)
                rc = upload(dif, buf, transaction)
                if rc < 0:
                    _logger.error("Error during upload")
                    ret = rc
                    break
                write_rc = file.file_p.write(buf[:rc])
                if write_rc < rc:
                    _logger.error(f"Short file write: {rc}")
                    ret = -1
                    break
                total_bytes += rc
                if rc < xfer_size or total_bytes >= upload_limit:
                    # Last block, return successfully
                    ret = total_bytes
                    break

                progress.update(description="Uploading...", advance=xfer_size)
                transaction += 1

            progress.update(description="Upload finished!")
            return ret
    except Exception as e:
        logger.error(e)
        return -1


def dnload_chunk(dif: dfu.DfuIf, data: bytes, size: int, transaction: int) -> int:
    """
    Download a chunk of data during DFU download operation.

    :param dif: DfuIf object representing the DFU interface
    :param data: Data to be downloaded
    :param size: Size of the data chunk
    :param transaction: Transaction number
    :return: Number of bytes sent or error code
    """

    ret = download(dif, data if size else None, transaction)

    if ret < 0:
        _logger.error("Error during download")
        return ret

    bytes_sent = ret

    while True:
        ret = int(status := dif.get_status())
        if ret < 0:
            _logger.error("Error during download get_status")
            return ret

        # dst = ret # useless?
        milli_sleep(status.bwPollTimeout)

        if status.bState in (dfu.State.DFU_DOWNLOAD_IDLE,
                             dfu.State.DFU_ERROR,
                             dfu.State.DFU_MANIFEST):
            break

    if status.bState == dfu.State.DFU_MANIFEST:
        _logger.info("Transitioning to dfuMANIFEST state")

    if status.bStatus != dfu.Status.OK:
        _logger.error("Download failed!")
        _logger.error("state(%u) = %s, status(%u) = %s", status.bState,
                      status.bState.to_string(), status.bStatus,
                      status.bStatus.to_string())
        return -1

    return bytes_sent


# Writes an element of any size to the device, taking care of page erases
# returns 0 on success, otherwise -EINVAL
# pylint: disable=invalid-name
def dnload_element(dif: dfu.DfuIf,
                   dw_element_address: int,
                   dw_element_size: int,
                   data: bytes,
                   xfer_size: int) -> int:
    """
    Download an element in DFU.

    :param dif: DfuIf object representing the DFU interface
    :param dw_element_address: Element address
    :param dw_element_size: Size of the element
    :param data: Data to be downloaded
    :param xfer_size: Transfer size
    :return: 0 if successful, error code otherwise
    """

    ret = 0
    segment = find_segment(MEM_LAYOUT, dw_element_address + dw_element_size - 1)

    if not segment or not segment.mem_type & DFUSE.WRITEABLE:
        _logger.error(
            f"Error: Last page at 0x{dw_element_address + dw_element_size - 1:08x} is not writeable"
        )
        return -1

    p = 0
    while p < dw_element_size:
        address = dw_element_address + p
        chunk_size = min(xfer_size, dw_element_size - p)

        segment = find_segment(MEM_LAYOUT, address)
        if not segment or not segment.mem_type & DFUSE.WRITEABLE:
            _logger.error(f"Error: Page at 0x{address:08x} is not writeable")
            return -1

        if VERBOSE:
            _logger.info(f"Download from image offset {p:08x} "
                        f"to memory {address:08x}-{address + chunk_size - 1:08x}"
                        f", size {chunk_size}")

        special_command(dif, address, Command.SET_ADDRESS)

        # transaction = 2 for no address offset
        ret = dnload_chunk(dif, data[p:p + chunk_size], chunk_size, 2)
        if ret != chunk_size:
            _logger.error(f"Failed to write whole chunk: {ret} of {chunk_size} bytes")
            return -1

        # Move to the next chunk
        p += xfer_size

    if not VERBOSE:
        _logger.info("")

    return ret


# Download raw binary file to DfuSe device
def do_bin_dnload(dif: dfu.DfuIf, xfer_size: int, file: DFUFile, start_address: int) -> int:
    """
    Download binary data to the specified address.

    :param dif: DfuIf object representing the DFU interface
    :param xfer_size: Transfer size
    :param file: DFUFile object containing the binary file
    :param start_address: Start address for the download
    :return: Number of bytes read or error code
    """
    dwElementAddress = start_address
    dwElementSize = file.size

    _logger.info(f"Downloading to address = 0x{dwElementAddress:08x}, size = {dwElementSize}")

    data = file.file_p.read()
    read_bytes = len(data)

    ret = dnload_element(dif, dwElementAddress, dwElementSize, data, xfer_size)
    if ret != 0:
        return ret

    if read_bytes != file.size:
        _logger.warning(f"Read {read_bytes} bytes, file size {file.size}")

    _logger.info("File downloaded successfully")
    return read_bytes


# Parse a DfuSe file and download contents to device
def do_dfuse_dnload(dif: dfu.DfuIf, xfer_size: int, file: DFUFile) -> int:
    """
    Download data from a DfuSe file to the DFU device.

    :param dif: DfuIf object representing the DFU interface
    :param xfer_size: Transfer size
    :param file: DFUFile object containing the DfuSe file
    :return: Number of bytes read or error code
    """
    dfu_prefix = file.file_p.read(11)
    read_bytes = len(dfu_prefix)

    if b'DfuSe\x01' not in dfu_prefix:
        _logger.error("No valid DfuSe signature")
        return -errno.EINVAL

    bTargets = dfu_prefix[10]
    _logger.info(f"File contains {bTargets} DFU images")

    for image in range(1, bTargets + 1):
        _logger.info(f"Parsing DFU image {image}")
        target_prefix = file.file_p.read(274)
        read_bytes += len(target_prefix)

        if target_prefix[:6] != b'Target':
            _logger.error("No valid target signature")
            return -errno.EINVAL

        bAlternateSetting = target_prefix[6]
        dwNbElements = int.from_bytes(target_prefix[266:270], byteorder='little')
        size = int.from_bytes(target_prefix[270:274], byteorder='little')
        _logger.info(
            f"Image for alternate setting {bAlternateSetting}, "
            f"({dwNbElements} elements, total size = {size})")

        if bAlternateSetting != dif.altsetting:
            _logger.warning("Image does not match current alternate setting.")
            _logger.warning("Please rerun with the correct -a option setting"
                           " to download this image!")

        for element in range(1, dwNbElements + 1):
            _logger.info(f"Parsing element {element}")
            element_header = file.file_p.read(8)
            dwElementAddress, dwElementSize, *_ = element_header
            _logger.info(f"Address = 0x{dwElementAddress:08x}, Size = {dwElementSize}")

            # Sanity check
            if read_bytes + dwElementSize + file.suffix_len > file.size:
                _logger.error("File too small for element size")
                return -errno.EINVAL

            data = file.file_p.read(dwElementSize)
            read_bytes += len(data)

            if bAlternateSetting == dif.altsetting:
                ret = dnload_element(dif, dwElementAddress, dwElementSize, data, xfer_size)
                if ret != 0:
                    return ret

    # Read through the whole file for bookkeeping
    file.file_p.read(file.suffix_len)
    read_bytes += file.suffix_len

    if read_bytes != file.size:
        _logger.warning(f"Read {read_bytes} bytes, file size {file.size}")

    _logger.info("Done parsing DfuSe file")
    return read_bytes


def do_dnload(dif: dfu.DfuIf, xfer_size: int, file: DFUFile, dfuse_options: [str, bytes]) -> int:
    """
    Perform DFU download operation.

    :param dif: DfuIf object representing the DFU interface
    :param xfer_size: Transfer size
    :param file: DFUFile object representing the file to be downloaded
    :param dfuse_options: DFU options string containing address, modifiers, and values
    :return: Number of bytes sent or error code
    """

    # pylint: disable=global-statement
    global MEM_LAYOUT
    ret: int

    try:
        if not dfuse_options:
            raise ValueError("No DFUse options provided")

        opts = parse_options(dfuse_options)

        MEM_LAYOUT = parse_memory_layout(dif.alt_name)  # HOTFIX
        if not MEM_LAYOUT:
            raise IOError("Failed to parse memory layout")

        if opts.unprotect:
            if not opts.force:
                raise PermissionError(
                    "The read unprotect command will erase the flash memory"
                    " and can only be used with force"
                )
            special_command(dif, 0, Command.READ_UNPROTECT)
            _logger.info("Device disconnects, erases flash and resets now")
            sys.exit(0)

        if opts.mass_erase:
            if not opts.force:
                raise PermissionError("The mass erase command can only be used with force")
            _logger.info("Performing mass erase, this can take a moment")
            special_command(dif, 0, Command.MASS_ERASE)

    except (ValueError, PermissionError, IOError) as err:
        _logger.error(err)
        raise GeneralError

    if opts.address:
        if file.bcdDFU == 0x11a:
            _logger.error("This is a DfuSe file, not meant for raw download")
            return -1
        ret = do_bin_dnload(dif, xfer_size, file, opts.address)
    else:
        if file.bcdDFU != 0x11a:
            _logger.error("Only DfuSe file version 1.1a is supported"
                         ", (for raw binary download, use the --dfuse-address option)")
            return -1
        ret = do_dfuse_dnload(dif, xfer_size, file)

    # free_segment_list(MEM_LAYOUT)
    MEM_LAYOUT = None

    if opts.leave:
        dnload_chunk(dif, b'', 0, 2)  # Zero-size
        # ret2 = int(dst := dfu.get_status(dif.dev, dif.interface))
        ret2 = int(dst := dif.get_status())
        if ret2 < 0:
            _logger.error("Error during download get_status")
        if VERBOSE:
            _logger.info(f"bState = {dst.bState} and bStatus = {dst.bStatus}")

    return ret



import unittest
from pydfuutil.__main__ import *


class TestMain(unittest.TestCase):

    @unittest.skip("Already implemented")
    def test_find_dfu_if_first(self):
        ctx = usb.core.find(find_all=True)
        for dev in ctx:
            dfu_if = next((i for i in find_dfu_if(dev) if i is not None), None)
            if dfu_if:
                print_dfu_if(dfu_if)

    @unittest.skip("Already implemented")
    def test_get_first_dfu_if(self):
        dev = usb.core.find(idVendor=0x1fc9, idProduct=0x000c)
        if dfu_if := get_first_dfu_if(dev):
            print_dfu_if(dfu_if)


    def test_usb_path2devnum(self):
        ret = usb_path2devnum('1-0.8:1.0')
        self.assertEqual(ret, 8)

    def test_get_first_in_gen(self):

        def gen():
            for i in range(5, 10):
                yield None
                yield i

        def gen2():
            for i in range(5, 10):
                yield None

        v = next((v for v in gen() if v is not None), None)
        v2 = next((v2 for v2 in gen2() if v2 is not None), None)
        self.assertEqual(v, 5)
        self.assertEqual(v2, None)

    @unittest.skip("Already implemented")
    def test_main_upload(self):
        argv = "-U test.bin -t 2048 -y".split(' ')
        main(argv)


if __name__ == '__main__':
    unittest.main()



"""
PyDfuUtil - Pure python fork of **[dfu-util](https://github.com/Stefan-Schmidt/dfu-util)**
wrappers to **[libusb](https://github.com/libusb/libusb)
"""

__author__ = "o-murphy"
__credits__ = ["Dmytro Yaroshenko"]
__copyright__ = ('2023 Yaroshenko Dmytro (https://github.com/o-murphy)',)



import logging
import unittest
from pydfuutil.quirks import *
from pydfuutil.logger import logger

logger.setLevel(logging.DEBUG)


class TestSetQuirks(unittest.TestCase):

    def test_quirk_polltimeout(self):
        self.assertEqual(set_quirks(VENDOR_OPENMOKO, 123, 1), QUIRK_POLLTIMEOUT)
        self.assertEqual(set_quirks(VENDOR_FIC, 456, 1), QUIRK_POLLTIMEOUT)
        self.assertEqual(set_quirks(VENDOR_VOTI, 789, 1), QUIRK_POLLTIMEOUT)

    def test_quirk_force_dfu11(self):
        self.assertEqual(set_quirks(VENDOR_LEAFLABS, PRODUCT_MAPLE3, 0x0200), QUIRK_FORCE_DFU11)
        self.assertEqual(set_quirks(VENDOR_LEAFLABS, PRODUCT_MAPLE3, 0x0100), 0)
        self.assertEqual(set_quirks(VENDOR_LEAFLABS, 123, 0x0200), 0)
        self.assertEqual(set_quirks(VENDOR_LEAFLABS, PRODUCT_MAPLE3, 0x0300), 0)

if __name__ == '__main__':
    unittest.main()



"""
Low-level DFU communication routines (part of dfu-programmer).
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

# import inspect
import sys
from dataclasses import dataclass
from enum import IntEnum, IntFlag

import usb.util

from pydfuutil.logger import logger
from pydfuutil.portable import milli_sleep
from pydfuutil.usb_dfu import FuncDescriptor

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])


class State(IntEnum):
    """Dfu states"""
    APP_IDLE = 0x00
    APP_DETACH = 0x01
    DFU_IDLE = 0x02
    DFU_DOWNLOAD_SYNC = 0x03
    DFU_DOWNLOAD_BUSY = 0x04
    DFU_DOWNLOAD_IDLE = 0x05
    DFU_MANIFEST_SYNC = 0x06
    DFU_MANIFEST = 0x07
    DFU_MANIFEST_WAIT_RESET = 0x08
    DFU_UPLOAD_IDLE = 0x09
    DFU_ERROR = 0x0a

    UNKNOWN_ERROR = -1

    def to_string(self):
        """
        :return: State.self name by State Enum
        """
        return state_to_string(self)





class Status(IntEnum):
    """Dfu statuses"""
    OK = 0x00
    ERROR_TARGET = 0x01
    ERROR_FILE = 0x02
    ERROR_WRITE = 0x03
    ERROR_ERASE = 0x04
    ERROR_CHECK_ERASED = 0x05
    ERROR_PROG = 0x06
    ERROR_VERIFY = 0x07
    ERROR_ADDRESS = 0x08
    ERROR_NOTDONE = 0x09
    ERROR_FIRMWARE = 0x0a
    ERROR_VENDOR = 0x0b
    ERROR_USBR = 0x0c
    ERROR_POR = 0x0d
    ERROR_UNKNOWN = 0x0e
    ERROR_STALLEDPKT = 0x0f

    def to_string(self):
        """
        :return: Status.self name by Status Enum
        """
        return status_to_string(self)


class Command(IntEnum):
    """Dfu commands"""
    DETACH = 0
    DNLOAD = 1
    UPLOAD = 2
    GETSTATUS = 3
    CLRSTATUS = 4
    GETSTATE = 5
    ABORT = 6


# /* DFU interface */
class Mode(IntFlag):
    """Dfu modes"""
    IFF_DFU = 0x0001  # /* DFU Mode, (not Runtime) */
    IFF_VENDOR = 0x0100
    IFF_PRODUCT = 0x0200
    IFF_CONFIG = 0x0400
    IFF_IFACE = 0x0800
    IFF_ALT = 0x1000
    IFF_DEVNUM = 0x2000
    IFF_PATH = 0x4000
    # DFU_IFF_DFU = 0x0001  /* DFU Mode, (not Runtime) */
    # DFU_IFF_ALT = 0x0002  /* Multiple alternate settings */


@dataclass(frozen=True)
class StatusRetVal:
    """
    Converts dfu_get_status result bytes to applicable dataclass
    This is based off of DFU_GETSTATUS
    the data structure to be populated with the results
    """
    # pylint: disable=invalid-name
    bStatus: Status = Status.ERROR_UNKNOWN
    bwPollTimeout: int = 0
    bState: State = State.DFU_ERROR
    iString: int = 0

    @classmethod
    def from_bytes(cls, data: bytes):
        """Creates StatusRetVal instance from bytes sequence"""
        if len(data) >= 6:
            bStatus = (Status(data[0])
                       if data[0] in Status.__members__.values()
                       else Status.ERROR_UNKNOWN)
            bwPollTimeout = int.from_bytes(data[1:4], 'little')
            bState = (State(data[4])
                      if data[0] in State.__members__.values()
                      else State.DFU_ERROR)
            iString = data[5]
            return cls(
                bStatus,
                bwPollTimeout,
                bState,
                iString
            )
        return cls()

    def __bytes__(self) -> bytes:
        return (
                bytes([self.bStatus.value])
                + self.bwPollTimeout.to_bytes(3, 'little')
                + bytes([self.bState.value, self.iString])
        )

    def __int__(self):
        return int.from_bytes(self, 'little')


@dataclass
class DfuIf:  # pylint: disable=too-many-instance-attributes
    """DFU Interface dataclass"""
    # pylint: disable=invalid-name
    vendor: int = None
    product: int = None
    bcdDevice: int = None
    configuration: int = None
    interface: int = None
    altsetting: int = None
    alt_name: str = None
    bus: int = None
    devnum: int = None
    path: [str, int] = None
    flags: [Mode, int] = 0
    count: int = None
    dev: usb.core.Device = None
    quirks: int = None
    bwPollTimeout: int = 0
    serial_name: str = ""
    usb_dfu_func_desc: FuncDescriptor = None


    @property
    def device_ids(self) -> dict:
        """Returns filter dict for usb.core.find() by VID:PID"""
        id_filter = {}
        if self.vendor:
            id_filter["idVendor"] = self.vendor
        if self.product:
            id_filter["idProduct"] = self.product
        return id_filter

    # The binds to direct dfu functions to get more pythonic
    # Use them better instead of direct

    def detach(self, timeout: int) -> bytes:
        """Binds self to dfu.detach()"""
        return detach(self.dev, self.interface, timeout)

    def download(self, transaction: int, data_or_length: [bytes, int]) -> int:
        """Binds self to dfu.download()"""
        return download(self.dev, self.interface, transaction, data_or_length)

    def upload(self, transaction: int, data_or_length: [bytes, int]) -> bytes:
        """Binds self to dfu.upload()"""
        return upload(self.dev, self.interface, transaction, data_or_length)

    def abort(self) -> int:
        """Binds self to dfu.abort()"""
        return abort(self.dev, self.interface)

    def get_status(self) -> StatusRetVal:
        """Binds self to dfu.get_status()"""
        return get_status(self.dev, self.interface)

    def get_state(self) -> State:
        """Binds self to dfu.get_state()"""
        return get_state(self.dev, self.interface)

    def abort_to_idle(self):
        """Binds self to dfu.abort_to_idle()"""
        return abort_to_idle(self)


def init(timeout: int) -> None:
    """
    Initiate dfu_util library with specified commands timeout
    :param timeout in milliseconds
    :return: None
    """

    # pylint: disable=global-statement
    global TIMEOUT

    if timeout > 0:
        TIMEOUT = timeout
    else:
        if 0 != DEBUG_LEVEL:
            raise ValueError(f"dfu_init: Invalid timeout value {timeout}")


# def verify_init() -> int:
#     """
#     Verifies provided TIMEOUT and DEBUG_LEVEL
#     NOTE: (function: typing.Callable) not needed cause python can get it from stack
#     :raise ValueError with caller function name
#     :return: 0
#     """
#     caller = inspect.stack()[0][3]
#     if INVALID_DFU_TIMEOUT == TIMEOUT:
#         if 0 != DEBUG_LEVEL:
#             raise ValueError(f'"{caller}": dfu system not initialized properly.')
#     return 0


def debug(level: int) -> None:
    """
    NOTE: Maybe not needed cause python can define globals after
    :param level: logging level (DEBUG, INFO, WARNING, ERROR)
    """

    # pylint: disable=global-statement
    global DEBUG_LEVEL
    DEBUG_LEVEL = level
    _logger.setLevel(level)


def detach(device: usb.core.Device, interface: int, timeout: int) -> bytes:
    """
    DETACH Request (DFU Spec 1.0, Section 5.1)

    Sends to device command to switch it to DFU mode
    u have to free device and handle it again
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :param timeout: the timeout in ms the USB device should wait for a pending
    :return: bytes or < 0 on error
    """
    # verify_init()
    _logger.debug('DETACH...')
    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_OUT
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.DETACH,
        wValue=timeout,
        wIndex=interface,
        data_or_wLength=None,
        timeout=TIMEOUT,
    )
    _logger.debug(f'DETACH {result >= 0}')
    return result


def download(device: usb.core.Device,
             interface: int,
             transaction: int,
             data_or_length: [bytes, int]) -> int:
    """
    DNLOAD Request (DFU Spec 1.0, Section 6.1.1)

    Download data to special page of DFU device
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :param transaction: start page int(total_data_size/xfer_size)
    :param data_or_length: the data to transfer
    :return: downloaded data or error code in bytes
    """
    # verify_init()
    _logger.debug('DFU_DOWNLOAD...')

    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_OUT
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.DNLOAD,
        wValue=transaction,
        wIndex=interface,
        data_or_wLength=data_or_length,
        timeout=TIMEOUT,
    )

    _logger.debug(f'DFU_DOWNLOAD {result >= 0}')
    return result


def upload(device: usb.core.Device,
           interface: int,
           transaction: int,
           data_or_length: [bytes, int]) -> bytes:
    """
    UPLOAD Request (DFU Spec 1.0, Section 6.2)

    Uploads data from special page of DFU device
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :param transaction: start page int(total_data_size/xfer_size)
    :param data_or_length: the buffer to put the received data in
    :return: uploaded bytes or < 0 on error
    """
    # verify_init()
    _logger.debug('UPLOAD...')

    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_IN
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.UPLOAD,
        wValue=transaction,
        wIndex=interface,
        data_or_wLength=data_or_length,
        timeout=TIMEOUT,
    )

    _logger.debug(f'UPLOAD {len(result) >= 0}')

    return result.tobytes()


def get_status(device: usb.core.Device, interface: int) -> StatusRetVal:
    """
     GETSTATUS Request (DFU Spec 1.0, Section 6.1.2)

    Returns DFU interface status
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :return: StatusRetVal
    """
    # verify_init()
    _logger.debug('DFU_GET_STATUS...')

    length = 6
    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_IN
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.GETSTATUS,
        wValue=0,
        wIndex=interface,
        data_or_wLength=length,
        timeout=TIMEOUT,
    )

    if len(result) == length:
        status = StatusRetVal.from_bytes(result.tobytes())
        _logger.debug(f'GET_STATUS {len(result) == 6}')
        _logger.debug(f'CURRENT STATE {status.bState.to_string()}')
        return status
    return StatusRetVal.from_bytes(result)


def clear_status(device: usb.core.Device, interface: int) -> int:
    """
    CLRSTATUS Request (DFU Spec 1.0, Section 6.1.3)

    Clears DFU interface status
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :return: return 0 or < 0 on an error
    """
    # verify_init()
    _logger.debug('CLEAR_STATUS...')

    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_OUT
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.CLRSTATUS,
        wValue=0,
        wIndex=interface,
        data_or_wLength=None,
        timeout=TIMEOUT,
    )
    _logger.debug(f'CLEAR_STATUS {result >= 0}')

    return result


def get_state(device: usb.core.Device, interface: int) -> [State, int]:
    """
    GETSTATE Request (DFU Spec 1.0, Section 6.1.5)

    Returns DFU interface state
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :return: returns the state or < 0 on error
    """
    # verify_init()

    length = 1
    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_IN
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.GETSTATE,
        wValue=0,
        wIndex=interface,
        data_or_wLength=length,
        timeout=TIMEOUT,
    )
    value = result.tobytes()[0] < 1
    if value < 1:
        return State(-1)
    if value in State.__members__.values():
        value = State(value)
    return value


def abort(device: usb.core.Device, interface: int) -> int:
    """
    ABORT Request (DFU Spec 1.0, Section 6.1.4)

    Aborts DFU command
    :param device: the usb_dev_handle to communicate with
    :param interface: the interface to communicate with
    :return: returns 0 or < 0 on an error
    """
    # verify_init()
    _logger.debug('ABORT...')

    result = device.ctrl_transfer(
        bmRequestType=usb.util.ENDPOINT_OUT
                      | usb.util.CTRL_TYPE_CLASS
                      | usb.util.CTRL_RECIPIENT_INTERFACE,
        bRequest=Command.ABORT,
        wValue=0,
        wIndex=interface,
        data_or_wLength=None,
        timeout=TIMEOUT,
    )

    _logger.debug(f'ABORT {result >= 0}')

    return result


def state_to_string(state: int) -> [str, None]:
    """
    :param state:
    :return: State name by State Enum
    """
    try:
        return _STATES_NAMES[State(state)]
    except (ValueError, KeyError):
        return None


def status_to_string(status: int) -> [str, None]:
    """
    :param status:
    :return: State name by Status Enum
    """
    try:
        return _DFU_STATUS_NAMES[Status(status)]
    except (ValueError, KeyError):
        return None


_STATES_NAMES = {
    State.APP_IDLE: 'appIDLE',
    State.APP_DETACH: 'appDETACH',
    State.DFU_IDLE: 'dfuIDLE',
    State.DFU_DOWNLOAD_SYNC: 'dfuDNLOAD-SYNC',
    State.DFU_DOWNLOAD_BUSY: 'dfuDNBUSY',
    State.DFU_DOWNLOAD_IDLE: 'dfuDNLOAD-IDLE',
    State.DFU_MANIFEST_SYNC: 'dfuMANIFEST-SYNC',
    State.DFU_MANIFEST: 'dfuMANIFEST',
    State.DFU_MANIFEST_WAIT_RESET: 'dfuMANIFEST-WAIT-RESET',
    State.DFU_UPLOAD_IDLE: 'dfuUPLOAD-IDLE',
    State.DFU_ERROR: 'dfuERROR',
}


_DFU_STATUS_NAMES = {
    Status.OK: "No error condition is present",
    Status.ERROR_TARGET: "File is not targeted for use by this device",
    Status.ERROR_FILE: "File is for this device but fails some vendor-specific test",
    Status.ERROR_WRITE: "Device is unable to write memory",
    Status.ERROR_ERASE: "Memory erase function failed",
    Status.ERROR_CHECK_ERASED: "Memory erase check failed",
    Status.ERROR_PROG: "Program memory function failed",
    Status.ERROR_VERIFY: "Programmed memory failed verification",
    Status.ERROR_ADDRESS: "Cannot program memory due to received address that is out of range",
    Status.ERROR_NOTDONE: "Received DNLOAD with wLength = 0, "
                          "but device does not think that it has all data yet",
    Status.ERROR_FIRMWARE: "Device's firmware is corrupt. "
                           "It cannot return to run-time (non-DFU) operations",
    Status.ERROR_VENDOR: "iString indicates a vendor specific error",
    Status.ERROR_USBR: "Device detected unexpected USB reset signalling",
    Status.ERROR_POR: "Device detected unexpected power on reset",
    Status.ERROR_UNKNOWN: "Something went wrong, but the device does not know what it was",
    Status.ERROR_STALLEDPKT: "Device stalled an unexpected request"
}


def abort_to_idle(dif: DfuIf):
    if dif.abort() < 0:
        _logger.error("Error sending dfu abort request")
        sys.exit(1)
    if (ret := int(dst := dif.get_status())) < 0:
        _logger.error("Error during abort get_status")
        sys.exit(1)
    if dst.bState != State.DFU_IDLE:
        _logger.error("Failed to enter idle state on abort")
        sys.exit(1)
    milli_sleep(dst.bwPollTimeout)
    return ret


# global definitions
DEBUG: int = 0

INVALID_DFU_TIMEOUT = -1

TIMEOUT: int = INVALID_DFU_TIMEOUT
TRANSACTION: int = 0

DEBUG_LEVEL: int = 0



"""Pydfuutil exceptions."""


class GeneralError(Exception):
    """
    Usually indicates a general error.
    This can be used to indicate that the program
    encountered some issue during execution
    that prevented it from completing its task successfully.
    It can also be used as a catch-all for unspecified errors.
    """
    exit_code = 1


class UsbIOError(GeneralError):
    exit_code = 1


class GeneralWarning(GeneralError):
    """
    Usually indicates a general warning
    """


class MisuseError(GeneralError):
    """
    Often used to indicate misuse or incorrect usage of the program.
    For example, if the program receives
    invalid command-line arguments or options,
    it might exit with code 2 to indicate a syntax error.
    """
    exit_code = 2

class CapabilityError(GeneralError):
    """DFU incompatible usage error."""
    exit_code = 3



import unittest
from pydfuutil import dfu
from unittest.mock import patch, MagicMock
import array


class TestDfu(unittest.TestCase):

    def setUp(self):
        dfu.init(1000)

    def test_init(self):
        self.assertEqual(dfu.TIMEOUT, 1000)

    def test_str(self):
        self.assertEqual(dfu.status_to_string(dfu.Status.OK), "No error condition is present")
        self.assertEqual(dfu.Status.OK.to_string(), "No error condition is present")
        self.assertEqual(dfu.state_to_string(dfu.State.APP_IDLE), "appIDLE")
        self.assertEqual(dfu.State.APP_IDLE.to_string(), "appIDLE")

    @patch("usb.core.find")
    def test_detach(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = 1

        # Calling the function
        interface = 0
        timeout = 1000
        result = dfu.detach(mock_device, interface, timeout)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0x21 | 0x01,
            bRequest=dfu.Command.DETACH,  # Assuming Command.DETACH is 23
            wValue=timeout,
            wIndex=interface,
            data_or_wLength=None,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(result, 1)  # Assuming success returns 1

    @patch("usb.core.find")
    def test_upload(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = array.array('B', [0] * 10)

        # Calling the function
        interface = 0
        transaction = 0
        data = bytes(10)
        result = dfu.upload(mock_device, interface, transaction, data)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0xa1,
            bRequest=dfu.Command.UPLOAD,  # Assuming Command.UPLOAD
            wValue=transaction,
            wIndex=interface,
            data_or_wLength=data,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(result, data)  # Assuming success returns bytes(10)

    @patch("usb.core.find")
    def test_dwnload(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = 10

        # Calling the function
        interface = 0
        transaction = 0
        data = bytes(10)
        result = dfu.download(mock_device, interface, transaction, data)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0x21,
            bRequest=dfu.Command.DNLOAD,  # Assuming Command.DNLOAD
            wValue=transaction,
            wIndex=interface,
            data_or_wLength=data,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(result, len(data))  # Assuming success returns 10

    @patch("usb.core.find")
    def test_get_status(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = array.array('B', [0] * 6)

        # Calling the function
        interface = 0
        transaction = 0
        length = 6
        status = dfu.get_status(mock_device, interface)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0xa1,
            bRequest=dfu.Command.GETSTATUS,  # Assuming Command.GETSTATUS
            wValue=transaction,
            wIndex=interface,
            data_or_wLength=length,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(status.bState, 0)  # Assuming success returns 0

    @patch("usb.core.find")
    def test_clear_status(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = 0

        # Calling the function
        interface = 0
        transaction = 0
        result = dfu.clear_status(mock_device, interface)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0x21,
            bRequest=dfu.Command.CLRSTATUS,  # Assuming Command.CLRSTATUS
            wValue=transaction,
            wIndex=interface,
            data_or_wLength=None,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(result, 0)  # Assuming success returns 0

    @patch("usb.core.find")
    def test_abort(self, mock_find):
        # Mocking the device
        mock_device = MagicMock()

        # Mocking the return value of usb.core.find
        mock_find.return_value = mock_device

        # Mocking the return value of ctrl_transfer method
        mock_device.ctrl_transfer.return_value = 0

        # Calling the function
        interface = 0
        transaction = 0
        result = dfu.abort(mock_device, interface)

        # Assertions
        mock_device.ctrl_transfer.assert_called_once_with(
            bmRequestType=0x21,
            bRequest=dfu.Command.ABORT,  # Assuming Command.ABORT
            wValue=transaction,
            wIndex=interface,
            data_or_wLength=None,
            timeout=1000,  # Assuming TIMEOUT is 1000
        )
        self.assertEqual(result, 0)  # Assuming success returns 0


if __name__ == "__main__":
    unittest.main()


"""
Simple quirk system for dfu-util
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)
"""

VENDOR_OPENMOKO = 0x1d50  # Openmoko Free-runner / GTA02
VENDOR_FIC = 0x1457  # Openmoko Free-runner / GTA02
VENDOR_VOTI = 0x16c0  # OpenPCD Reader
VENDOR_LEAFLABS = 0x1eaf  # Maple
PRODUCT_MAPLE3 = 0x0003  # rev 3 and 5

QUIRK_POLLTIMEOUT = 1 << 0
QUIRK_FORCE_DFU11 = 1 << 1

# Fallback value, works for OpenMoko
DEFAULT_POLLTIMEOUT = 5


# pylint: disable=invalid-name
def set_quirks(vendor: int, product: int, bcdDevice: int) -> int:
    """

    :param vendor: VID
    :param product: PID
    :param bcdDevice:
    :return: quirks
    """
    quirks = 0

    # Device returns bogus bwPollTimeout values
    if vendor in (VENDOR_OPENMOKO, VENDOR_FIC, VENDOR_VOTI):
        quirks |= QUIRK_POLLTIMEOUT

    # Reports wrong DFU version in DFU descriptor
    if vendor == VENDOR_LEAFLABS and product == PRODUCT_MAPLE3 and bcdDevice == 0x0200:
        quirks |= QUIRK_FORCE_DFU11

    return quirks



import unittest
from unittest.mock import Mock, patch

import usb.util

from pydfuutil import dfuse, dfu, dfuse_mem
from pydfuutil.dfu_file import DFUFile
from pydfuutil.dfuse_mem import parse_memory_layout


class TestSpecialCommand(unittest.TestCase):

    def setUp(self):
        dfuse.MEM_LAYOUT = parse_memory_layout(
            "@Internal Flash/0x08000000/04*016Kg,01*064Kg,01*128Kg"
        )
        self.dfu_if = Mock()
        self.dfu_if.abort.return_value = 0
        self.dfu_if.dev.ctrl_transfer.return_value = 0
        self.dfu_if.get_status.side_effect = [
            dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_BUSY, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
        ]
        self.address = 0x8000000

    def test_erase_page_command(self):
        ret = dfuse.special_command(self.dfu_if, self.address, dfuse.Command.ERASE_PAGE)
        self.assertGreater(ret, 0)

    def test_set_address_command(self):
        ret = dfuse.special_command(self.dfu_if, self.address, dfuse.Command.SET_ADDRESS)
        self.assertGreater(ret, 1)

    def test_mass_erase_command(self):
        ret = dfuse.special_command(self.dfu_if, self.address, dfuse.Command.MASS_ERASE)
        self.assertGreater(ret, 1)

    def test_read_unprotect_command(self):
        ret = dfuse.special_command(self.dfu_if, self.address, dfuse.Command.READ_UNPROTECT)
        self.assertGreater(ret, 1)


class TestUpload(unittest.TestCase):

    def setUp(self):
        dfu.init(5000)
        self.dfu_if = Mock()

    def test_upload_success(self):
        self.dfu_if.dev.ctrl_transfer.return_value = 1
        data = b'\x01\x02\x03\x04'
        transaction = 123
        status = dfuse.upload(self.dfu_if, data, transaction)
        self.assertEqual(status, 1)
        self.dfu_if.dev.ctrl_transfer.assert_called_once_with(
            bmRequestType=usb.util.ENDPOINT_IN | usb.util.CTRL_TYPE_CLASS | usb.util.CTRL_RECIPIENT_INTERFACE,
            bRequest=dfu.Command.UPLOAD,
            wValue=transaction,
            wIndex=self.dfu_if.interface,
            data_or_wLength=data,
            timeout=dfu.TIMEOUT
        )

    def test_upload_failure(self):
        self.dfu_if.dev.ctrl_transfer.return_value = -1
        data = b'\x01\x02\x03\x04'
        transaction = 123
        status = dfuse.upload(self.dfu_if, data, transaction)
        self.assertEqual(status, -1)
        self.dfu_if.dev.ctrl_transfer.assert_called_once()


class TestDownload(unittest.TestCase):

    def setUp(self):
        dfu.init(5000)
        self.dfu_if = Mock()

    def test_download_success(self):
        self.dfu_if.dev.ctrl_transfer.return_value = 1
        data = b'\x01\x02\x03\x04'
        transaction = 123
        status = dfuse.download(self.dfu_if, data, transaction)
        self.assertEqual(status, 1)
        self.dfu_if.dev.ctrl_transfer.assert_called_once_with(
            bmRequestType=usb.util.ENDPOINT_OUT | usb.util.CTRL_TYPE_CLASS | usb.util.CTRL_RECIPIENT_INTERFACE,
            bRequest=dfu.Command.DNLOAD,
            wValue=transaction,
            wIndex=self.dfu_if.interface,
            data_or_wLength=data,
            timeout=dfu.TIMEOUT
        )

    def test_download_failure(self):
        self.dfu_if.dev.ctrl_transfer.return_value = -1
        data = b'\x01\x02\x03\x04'
        transaction = 123
        status = dfuse.download(self.dfu_if, data, transaction)
        self.assertEqual(status, -1)
        self.dfu_if.dev.ctrl_transfer.assert_called_once()


class TestDnloadChunk(unittest.TestCase):
    def setUp(self):
        dfu.init(5000)
        self.dfu_if = Mock()

    @patch('pydfuutil.dfuse.download')
    def test_dnload_chunk_success(self, mock_download):
        mock_download.return_value = 4
        status_mock = dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_IDLE, bStatus=dfu.Status.OK)
        self.dfu_if.get_status.return_value = status_mock
        data = b'\x01\x02\x03\x04'
        size = len(data)
        transaction = 123
        bytes_sent = dfuse.dnload_chunk(self.dfu_if, data, size, transaction)
        self.assertEqual(bytes_sent, 4)
        mock_download.assert_called_once_with(self.dfu_if, data, transaction)
        self.assertEqual(self.dfu_if.get_status.call_count, 1)

    @patch('pydfuutil.dfuse.download')
    def test_dnload_chunk_failure(self, mock_download):
        mock_download.return_value = -1
        data = b'\x01\x02\x03\x04'
        size = len(data)
        transaction = 123
        bytes_sent = dfuse.dnload_chunk(self.dfu_if, data, size, transaction)
        self.assertEqual(bytes_sent, -1)
        mock_download.assert_called_once_with(self.dfu_if, data, transaction)
        self.assertEqual(self.dfu_if.get_status.call_count, 0)


class TestDoUpload(unittest.TestCase):

    def setUp(self):
        self.dfu_if = Mock()
        self.dfu_if.abort.return_value = 0
        self.dfu_if.alt_name = "@Internal Flash/0x08000000/04*016Kg,01*064Kg,01*128Kg"
        self.dfu_if.dev.ctrl_transfer.return_value = 0
        self.dfu_if.get_status.side_effect = [
            dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_BUSY, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
        ]
        self.dfuse_options = "--address 0x08000000"
        self.file = DFUFile("")
        self.file.file_p = Mock()
        self.chunk_sizes = [
            1024, 512, 256
        ]
        self.file.file_p.write.side_effect = self.chunk_sizes
        self.xfer_size = sum(self.chunk_sizes)

    @patch('pydfuutil.dfuse.parse_options')
    @patch('pydfuutil.dfuse.upload')
    def test_do_upload_with_options_and_address(self, mock_upload, mock_parse_options):
        mock_parse_options.return_value = Mock(address=0x08000000, length=0x4000, force=False)
        mock_upload.side_effect = self.chunk_sizes
        self.assertEqual(dfuse.do_upload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), 1024)

    @patch('pydfuutil.dfuse.parse_options')
    @patch('pydfuutil.dfuse.upload')
    def test_do_upload_with_options_and_no_address(self, mock_upload, mock_parse_options):
        mock_parse_options.return_value = Mock(address=None, length=0x4000, force=False)
        mock_upload.side_effect = self.chunk_sizes
        self.assertEqual(dfuse.do_upload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), 1024)

    @patch('pydfuutil.dfuse.parse_options')
    def test_do_upload_no_options(self, mock_parse_options):
        mock_parse_options.side_effect = ValueError("No options provided")
        self.assertEqual(dfuse.do_upload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), -1)


class TestDnloadElement(unittest.TestCase):

    def setUp(self):
        dfuse.MEM_LAYOUT = parse_memory_layout(
            "@Internal Flash/0x08000000/04*016Kg,01*064Kg,01*128Kg"
        )
        dfu.init(5000)
        self.dfu_if = Mock()
        self.dfu_if.abort.return_value = 0
        self.dfu_if.dev.ctrl_transfer.return_value = 0

        self.dw_element_address = 0x08000000
        self.dw_element_size = 1024
        self.data = b'\x01\x02\x03\x04'
        self.xfer_size = 512

        self.dfu_if.get_status.side_effect = [
                                                 dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_BUSY,
                                                                  bStatus=dfu.Status.OK),
                                                 dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
                                                 dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
                                             ] * (self.dw_element_size // self.xfer_size)

    @patch('pydfuutil.dfuse_mem.find_segment')
    @patch('pydfuutil.dfuse.dnload_chunk')
    def test_dnload_element_success(self, mock_dnload_chunk, mock_find_segment):
        mock_find_segment.return_value = Mock(mem_type=dfuse_mem.DFUSE.WRITEABLE)
        mock_dnload_chunk.return_value = 512
        self.assertEqual(
            dfuse.dnload_element(self.dfu_if, self.dw_element_address, self.dw_element_size, self.data, self.xfer_size),
            512)

    @patch('pydfuutil.dfuse_mem.find_segment')
    @patch('pydfuutil.dfuse.dnload_chunk')
    def test_dnload_element_segment_not_found(self, mock_dnload_chunk, mock_find_segment):
        mock_find_segment.return_value = None
        mock_dnload_chunk.return_value = -1
        self.assertEqual(
            dfuse.dnload_element(self.dfu_if, self.dw_element_address, self.dw_element_size, self.data, self.xfer_size),
            -1)

    @patch('pydfuutil.dfuse_mem.find_segment')
    @patch('pydfuutil.dfuse.dnload_chunk')
    def test_dnload_element_segment_not_writeable(self, mock_dnload_chunk, mock_find_segment):
        mock_find_segment.return_value = Mock(mem_type=0)
        mock_dnload_chunk.return_value = -1
        self.assertEqual(
            dfuse.dnload_element(self.dfu_if, self.dw_element_address, self.dw_element_size, self.data, self.xfer_size),
            -1)


class TestDoBinDnload(unittest.TestCase):

    def setUp(self):
        self.dfu_if = Mock()
        self.xfer_size = 1024
        self.start_address = 0x08000000
        self.file = Mock(spec=DFUFile)
        self.file.size = 1024
        self.file.file_p.read.return_value = bytes(1024)

    @patch('pydfuutil.dfuse.dnload_element')
    def test_do_bin_dnload_success(self, mock_dnload_element):
        mock_dnload_element.return_value = 0
        self.assertEqual(dfuse.do_bin_dnload(self.dfu_if, self.xfer_size, self.file, self.start_address),
                         self.file.size)

    @patch('pydfuutil.dfuse.dnload_element')
    def test_do_bin_dnload_failure(self, mock_dnload_element):
        mock_dnload_element.return_value = -1
        self.assertEqual(dfuse.do_bin_dnload(self.dfu_if, self.xfer_size, self.file, self.start_address), -1)


class TestDoDfuseDnload(unittest.TestCase):

    def setUp(self):
        self.dfu_if = Mock()
        self.xfer_size = 1024
        self.file = DFUFile("")
        self.file.file_p = Mock()
        self.file.suffix_len = 16

    @patch('pydfuutil.dfuse.dnload_element')
    def test_do_dfuse_dnload_success(self, mock_dnload_element):
        mock_dnload_element.return_value = 0
        # self.file.file_p.read.return_value = b'DfuSe\x01\x01Target\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        read_values = [
            b'DfuSe\x01\x01\x00\x00\x00\x00',
            bytes(16),
        ]
        self.file.file_p.read.side_effect = read_values
        self.file.size = 11 + 16
        self.assertEqual(dfuse.do_dfuse_dnload(self.dfu_if, self.xfer_size, self.file), self.file.size)

    def test_do_dfuse_dnload_invalid_signature(self):
        self.file.file_p.read.return_value = b'InvalidSignature'
        self.assertEqual(dfuse.do_dfuse_dnload(self.dfu_if, self.xfer_size, self.file), -22)  # -errno.EINVAL

    def test_do_dfuse_dnload_invalid_target_signature(self):
        self.file.file_p.read.return_value = b'DfuSe\x01\x00InvalidTrgt'
        self.assertEqual(dfuse.do_dfuse_dnload(self.dfu_if, self.xfer_size, self.file), -22)  # -errno.EINVAL


class TestDoDnload(unittest.TestCase):

    def setUp(self):
        self.dfu_if = Mock()
        self.dfu_if.alt_name = "@Internal Flash/0x08000000/04*016Kg,01*064Kg,01*128Kg"
        self.dfu_if.dev.ctrl_transfer.return_value = 0
        self.dfu_if.get_status.side_effect = (
            dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_BUSY, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_DOWNLOAD_IDLE, bStatus=dfu.Status.OK),
            dfu.StatusRetVal(bState=dfu.State.DFU_IDLE, bStatus=dfu.Status.OK),
        )
        self.xfer_size = 1024
        self.file = DFUFile("")
        self.file.file_p = Mock()
        self.file.file_p.read.side_effect = [bytes(1024), b'']
        # self.file.bcdDFU = 0x11a
        self.file.bcdDFU = 0x0100
        self.dfuse_options = ['--dfuse-address', b'\x00\x00\x00\x00']

    @patch('pydfuutil.dfuse.do_bin_dnload')
    @patch('pydfuutil.dfuse.parse_memory_layout')
    @patch('pydfuutil.dfuse.parse_options')
    def test_do_dnload_binary(self, mock_parse_options, mock_parse_memory_layout, mock_do_bin_dnload):
        mock_parse_options.return_value = Mock(address=0x1000, unprotect=False, mass_erase=False, leave=False)
        mock_do_bin_dnload.return_value = 1024
        self.assertEqual(dfuse.do_dnload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), 1024)
        mock_do_bin_dnload.assert_called_once_with(self.dfu_if, self.xfer_size, self.file, 0x1000)

    @patch('pydfuutil.dfuse.do_dfuse_dnload')
    @patch('pydfuutil.dfuse.parse_options')
    def test_do_dnload_dfuse(self, mock_parse_options, mock_do_dfuse_dnload):
        self.file.bcdDFU = 0x11a
        opts = Mock(address=0x08000000, force=True,
                    unprotect=None, mass_erase=False)
        mock_parse_options.return_value = opts
        mock_do_dfuse_dnload.return_value = 1024
        self.assertEqual(dfuse.do_dfuse_dnload(self.dfu_if, self.xfer_size, self.file), 1024)
        mock_do_dfuse_dnload.assert_called_once_with(self.dfu_if, self.xfer_size, self.file)

    @patch('pydfuutil.dfuse.parse_options')
    def test_do_dnload_invalid_binary(self, mock_parse_options):
        self.file.bcdDFU = 0x100
        self.file.file_p.read.return_value = bytes(20)
        opts = Mock(address=0x08000000, force=True,
                    unprotect=None, mass_erase=False)
        mock_parse_options.return_value = opts
        self.assertEqual(dfuse.do_dnload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), -1)

    @patch('pydfuutil.dfuse.parse_options')
    def test_do_dnload_invalid_dfuse(self, mock_parse_options):
        self.file.bcdDFU = 0x0100
        self.file.file_p.read.return_value = bytes(20)
        opts = Mock(address=0x08000000, force=True,
                    unprotect=None, mass_erase=False)
        mock_parse_options.return_value = opts
        self.assertEqual(dfuse.do_dnload(self.dfu_if, self.xfer_size, self.file, self.dfuse_options), -1)


if __name__ == '__main__':
    unittest.main()



"""
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)
"""

import argparse
import importlib.metadata
import os
import sys
from enum import IntEnum

from pydfuutil import __copyright__
from pydfuutil import lmdfu
from pydfuutil.dfu_file import DFUFile, parse_dfu_suffix
from pydfuutil.exceptions import GeneralWarning, GeneralError, MisuseError
from pydfuutil.logger import logger

try:
    __version__ = importlib.metadata.version("pydfuutil")
except importlib.metadata.PackageNotFoundError:
    __version__ = 'UNKNOWN'

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])


class Mode(IntEnum):
    """DFU suffix operate mode"""
    NONE = 0x1
    ADD = 0x2
    DEL = 0x3
    CHECK = 0x4


class LmdfuMode(IntEnum):
    """LMDFU suffix operate mode"""
    NONE = 0x1
    ADD = 0x2
    DEL = 0x3
    CHECK = 0x4


VERSION = (f'pydfuutil-suffix " {__version__} "\n {__copyright__[0]}\n'
           f'This program is Free Software and has ABSOLUTELY NO WARRANTY\n\n')


def check_suffix(file: DFUFile) -> int:
    """check dfu suffix for valid and prints it"""
    ret: int

    ret = parse_dfu_suffix(file)
    if ret > 0:
        print(f"The file {file.name} contains a DFU suffix with the following properties:")
        print(f"BCD device:\t0x{file.bcdDevice:04X}")
        print(f"Product ID:\t0x{file.idProduct:04X}")
        print(f"Vendor ID:\t0x{file.idVendor:04X}")
        print(f"BCD DFU:\t0x{file.bcdDFU:04X}")
        print(f"Length:\t\t{file.suffix_len}")
        print(f"CRC:\t\t0x{file.dwCRC:08X}")
    return ret


def remove_suffix(file: DFUFile) -> int:
    """remove suffix from DFU file"""
    ret: int

    ret = parse_dfu_suffix(file)
    if ret <= 0:
        return 0

    if hasattr(os, 'ftruncate'):
        try:
            with open(file.name, 'r+', encoding='utf-8') as f:
                f.truncate(file.size - file.suffix_len)
            _logger.info("DFU suffix removed")
        except OSError as e:
            raise GeneralError(f"Error truncating file: {e}")
    else:
        _logger.error("Suffix removal not implemented on this platform")
    return 1


def add_suffix(file: DFUFile, pid: int, vid: int, did: int) -> None:
    """Add suffix to DFU file"""
    ret: int

    file.idProduct = pid
    file.idVendor = vid
    file.bcdDevice = did

    ret = lmdfu.generate_dfu_suffix(file)
    if ret < 0:
        try:
            raise OSError("generate")
        except OSError as e:
            raise GeneralError(e)
    _logger.info("New DFU suffix added.")


def add_cli_options(parser: argparse.ArgumentParser) -> None:
    """Add cli options"""
    parser.add_argument('-V', '--version', action='version',
                        version=VERSION,
                        help='Print the version number')

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument('-c', '--check',
                       const=Mode.CHECK, dest='mode', action='store_const',
                       help='Check DFU suffix of <file>')
    group.add_argument('-a', '--add',
                       const=Mode.ADD, dest='mode', action='store_const',
                       help='Add DFU suffix to <file>')
    group.add_argument('-D', '--delete',
                       const=Mode.DEL, dest='mode', action='store_const',
                       help='Delete DFU suffix from <file>')

    parser.add_argument('file', action='store', metavar='<file>',
                        type=argparse.FileType('r+b'), default=None,
                        help="Target filename")

    parser.add_argument('-p', '--pid', action='store', metavar="<productID>",
                        required=False,
                        type=lambda x: int(x, 16), help='Add product ID into DFU suffix in <file>')
    parser.add_argument('-v', '--vid', action='store', metavar="<vendorID>",
                        required=False,
                        type=lambda x: int(x, 16), help='Add vendor ID into DFU suffix in <file>')
    parser.add_argument('-d', '--did', action='store', metavar="<deviceID>",
                        required=False,
                        type=lambda x: int(x, 16), help='Add device ID into DFU suffix in <file>')
    parser.add_argument('-s', '--stellaris-address', dest='lmdfu_flash_address',
                        metavar="<address>", type=int, help='Specify lmdfu address for LMDFU_ADD')
    parser.add_argument('-T', '--stellaris', dest='lmdfu_mode', action='store_const',
                        const=LmdfuMode.CHECK, help='Set lmdfu mode to LMDFU_CHECK')


def main() -> None:
    """cli entry point for suffix"""

    parser = argparse.ArgumentParser(
        prog='pydfuutil-suffix',
        exit_on_error=False,
    )
    add_cli_options(parser)
    print(f"v{__version__}")
    try:
        args = parser.parse_args()
    except argparse.ArgumentError as err:
        parser.print_help()
        raise GeneralError(err)

    lmdfu_mode = LmdfuMode.NONE
    lmdfu_flash_address: int = 0
    lmdfu_prefix: int = 0

    empty = 0xffff

    file = DFUFile(args.file.name)
    file.file_p, mode = args.file, args.mode

    pid = args.pid if args.pid else empty
    vid = args.vid if args.vid else empty
    did = args.did if args.did else empty

    if args.lmdfu_flash_address:
        lmdfu_mode, lmdfu_flash_address = LmdfuMode.ADD, args.lmdfu_flash_address

    if args.lmdfu_mode:
        lmdfu_mode = LmdfuMode.CHECK

    if mode == Mode.DEL and lmdfu_mode == LmdfuMode.CHECK:
        lmdfu_mode = LmdfuMode.DEL

    if mode != Mode.NONE:
        try:
            if mode == Mode.ADD:

                if check_suffix(file):
                    if lmdfu_prefix:
                        lmdfu.check_prefix(file)
                    raise GeneralWarning("Please remove existing DFU suffix before adding a new one.")

                if lmdfu_mode == LmdfuMode.ADD:
                    if lmdfu.check_prefix(file):
                        _logger.info("Adding new anyway")
                    lmdfu.add_prefix(file, lmdfu_flash_address)

                add_suffix(file, pid, vid, did)

            elif mode == Mode.CHECK:
                # Note: could open read-only here
                check_suffix(file)
                if lmdfu_mode == LmdfuMode.CHECK:
                    lmdfu.check_prefix(file)

            elif mode == Mode.DEL:
                if (not remove_suffix(file)
                        and lmdfu_mode == LmdfuMode.DEL
                        and lmdfu.check_prefix(file)):
                    lmdfu.remove_prefix(file)
                    raise GeneralWarning

            else:
                parser.print_help()
                raise MisuseError

            if lmdfu_mode == LmdfuMode.DEL and check_suffix(file):
                raise GeneralWarning(
                    "DFU suffix exist. "
                    "Remove suffix before using -T or use it with -D to delete suffix")

            if lmdfu_mode == LmdfuMode.DEL and lmdfu.check_prefix(file):
                lmdfu.remove_prefix(file)

        except Exception as error:
            raise GeneralError(error)


if __name__ == '__main__':
    try:
        main()
    except GeneralWarning as warn:
        if warn.__str__():
            _logger.warning(warn)
    except GeneralError as err:
        if err.__str__():
            _logger.error(err)
        sys.exit(err.exit_code)
    except Exception as err:
        logger.error(err)
        sys.exit(1)
    sys.exit(0)



"""
Default logger initializer for pydfuutil
"""

import logging

__all__ = ('logger', )

logging.basicConfig(level=logging.INFO)

formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger = logging.getLogger('pydfuutil')
logger.setLevel(logging.INFO)
logger.propagate = False
logger.addHandler(stream_handler)



"""
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)
"""

from time import sleep
__all__ = ("milli_sleep", )


def milli_sleep(msec: int) -> None:
    """
    :param msec: sleep timeout in milliseconds
    :return: None
    """
    sleep(int(msec / 1000))



"""
Checks for, parses and generates a DFU suffix
Load or store DFU files including suffix and prefix
(C) 2023 Yaroshenko Dmytro (https://github.com/o-murphy)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import io
import os
import struct
import sys
import warnings
from dataclasses import dataclass, field
from enum import Enum

from construct import (Struct, Const, ByteSwapped, Default,
                       Int32ub, Int16ub, Int8sb,
                       ConstError, StreamError)

from pydfuutil.logger import logger

__all__ = ('DFUFile', 'parse_dfu_suffix', 'generate_dfu_suffix')

_logger = logger.getChild(__name__.rsplit('.', maxsplit=1)[-1])

DFU_SUFFIX_LENGTH = 16
LMDFU_PREFIX_LENGTH = 8
LPCDFU_PREFIX_LENGTH = 16
STDIN_CHUNK_SIZE = 65536

crc32_table = [
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
    0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
    0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
    0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
    0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
    0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
    0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa, 0x42b2986c,
    0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
    0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
    0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
    0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190, 0x01db7106,
    0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
    0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
    0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
    0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
    0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
    0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
    0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
    0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
    0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
    0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
    0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
    0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
    0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
    0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
    0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
    0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
    0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
    0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
    0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
    0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
    0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
    0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
    0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
    0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
    0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
    0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
    0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
    0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
    0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
    0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
]

_suffix = ByteSwapped(Struct(
    dwCRC=Default(Int32ub, 0xffffffff),
    bLength=Const(DFU_SUFFIX_LENGTH, Int8sb),
    ucDfuSignature=Const(b'DFU'),
    bcdDFU=Int16ub,
    idVendor=Int16ub,
    idProduct=Int16ub,
    bcdDevice=Int16ub,
))


@dataclass
class DFUFile:  # pylint: disable=too-many-instance-attributes, invalid-name
    """Class to store DFU file data"""
    name: [str, None]
    file_p: io.FileIO = None
    size: int = 0
    dwCRC: int = 0
    suffix_len: int = 0
    bcdDFU: int = 0
    idVendor: int = 0xffff  # wildcard value
    idProduct: int = 0xffff  # wildcard value
    bcdDevice: int = 0xffff  # wildcard value

    def parse_dfu_suffix(self) -> int:
        """Bind parse_dfu_suffix to DFUFile instance"""
        return parse_dfu_suffix(self)

    def generate_dfu_suffix(self) -> int:
        """Bind generate_dfu_suffix to DFUFile instance"""
        return generate_dfu_suffix(self)


@dataclass
class DFUFileSize:
    total: int
    prefix: int
    suffix: int


class SuffixReq(Enum):
    NO_SUFFIX = 0
    NEEDS_SUFFIX = 1
    MAYBE_SUFFIX = 2


class PrefixReq(Enum):
    NO_SUFFIX = 0
    NEEDS_SUFFIX = 1
    MAYBE_SUFFIX = 2


class PrefixType(Enum):
    ZERO_PREFIX = 0
    LMDFU_PREFIX = 1
    LPCDFU_UNENCRYPTED_PREFIX = 2


@dataclass
class DFUFile011:  # pylint: disable=too-many-instance-attributes, invalid-name
    """Class to store DFU file data"""
    name: [str, None]
    firmware: [bytearray, bytes] = field(default_factory=bytearray)
    file_p: io.FileIO = None
    size: DFUFileSize = None
    lmdfu_address: int = 0
    prefix_type: PrefixType = None
    dwCRC: int = 0
    bcdDFU: int = 0
    idVendor: int = 0xffff  # wildcard value
    idProduct: int = 0xffff  # wildcard value
    bcdDevice: int = 0xffff  # wildcard value

    def parse_dfu_suffix(self) -> int:
        """Bind parse_dfu_suffix to DFUFile instance"""
        return parse_dfu_suffix(self)

    def generate_dfu_suffix(self) -> int:
        """Bind generate_dfu_suffix to DFUFile instance"""
        return generate_dfu_suffix(self)


def crc32_byte(accum: int, delta: int):
    """
    Calculate a 32-bit CRC
    """
    return crc32_table[(accum ^ delta) & 0xff] ^ (accum >> 8)


def probe_prefix(file: DFUFile011):
    prefix = file.firmware

    if file.size.total < LMDFU_PREFIX_LENGTH:
        return 1
    if prefix[0] == 0x01 and prefix[1] == 0x00:
        payload_len = ((prefix[7] << 24) | (prefix[6] << 16)
                       | (prefix[5] << 8) | prefix[4])
        expected_payload_len = file.size.total - LMDFU_PREFIX_LENGTH - file.size.suffix
        if payload_len != expected_payload_len:
            return 1
        file.prefix_type = PrefixType.LMDFU_PREFIX
        file.size.prefix = LMDFU_PREFIX_LENGTH
        file.lmdfu_address = 1024 * ((prefix[3] << 8) | prefix[2])

    elif ((prefix[0] & 0x3f) == 0x1a) and ((prefix[1] & 0x3f) == 0x3f):
        file.prefix_type = PrefixType.LPCDFU_UNENCRYPTED_PREFIX
        file.size.prefix = LPCDFU_PREFIX_LENGTH
    if file.size.prefix + file.size.suffix > file.size.total:
        return 1
    return 0


def write_crc(f: DFUFile.file_p, crc: int, buf: [bytes, bytearray], size: int) -> int:
    # compute CRC
    for x in range(0, size):
        crc = crc32_byte(crc, buf[x])

    # write data
    if f.write(buf, size) != size:
        _logger.error(f"Could not write {size} bytes to {f}")

    return crc


def load_file(file: DFUFile011, check_suffix: SuffixReq, check_prefix: PrefixReq):

    file.size.prefix = 0
    file.size.suffix = 0

    # default values, if no valid suffix is found
    file.bcdDFU = 0
    file.idVendor = 0xffff  # wildcard value
    file.idProduct = 0xffff  # wildcard value
    file.bcdDevice = 0xffff  # wildcard value

    # default values, if no valid prefix is found
    file.lmdfu_address = 0

    file.firmware = None

    if file.name and file.name != '-':

        if sys.stdin.isatty():
            print("Please provide input via stdin.")
            return None

        # Check if the platform is Windows
        if sys.platform.startswith("win"):
            import msvcrt
            # Set stdin to binary mode
            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)

        file.firmware = bytearray()
        read_bytes = sys.stdin.buffer.read(STDIN_CHUNK_SIZE)
        file.firmware += read_bytes
        file.size.total = len(read_bytes)

        while len(read_bytes) == STDIN_CHUNK_SIZE:
            read_bytes = sys.stdin.buffer.read(STDIN_CHUNK_SIZE)
            file.firmware += read_bytes
            file.size.total += len(read_bytes)

        _logger.debug(f"Read {file.size.total} bytes from stdin")

        # Never require suffix when reading from stdin
        check_suffix = SuffixReq.MAYBE_SUFFIX
    else:
        with open(file.name, 'rb') as file.file_p:
            file.firmware = file.file_p.read()
            file.size.total = len(file.firmware)

    # TODO
    raise NotImplementedError
    # # Check for possible DFU file suffix by trying to parse one
    # if check_suffix != SuffixReq.NO_SUFFIX:
    #     dfusuffix = file.firmware[-16:] if file.size['total'] >= 16 else None
    #     missing_suffix = False
    #     reason = None
    #
    #     if not dfusuffix:
    #         reason = "File too short for DFU suffix"
    #         missing_suffix = True
    #     elif dfusuffix[10:13] != b'DFU':
    #         reason = "Invalid DFU suffix signature"
    #         missing_suffix = True
    #     else:
    #         crc = 0xffffffff
    #         for byte in file.firmware[:-16]:
    #             crc = crc32_byte(crc, byte)
    #
    #         file.dwCRC = struct.unpack('<I', dfusuffix[12:16])[0]
    #         if file.dwCRC != crc:
    #             reason = "DFU suffix CRC does not match"
    #             missing_suffix = True
    #         else:
    #             file.bcdDFU = struct.unpack('<H', dfusuffix[6:8])[0]
    #             _logger.debug(f"DFU suffix version {file.bcdDFU}")
    #
    #             file.size.suffix = dfusuffix[11]
    #             if file.size.suffix < 16:
    #                 raise ValueError("Unsupported DFU suffix length")
    #             if file.size.suffix > file.size.total:
    #                 raise ValueError("Invalid DFU suffix length")
    #
    #             file.idVendor = struct.unpack('<H', dfusuffix[4:6])[0]
    #             file.idProduct = struct.unpack('<H', dfusuffix[2:4])[0]
    #             file.bcdDevice = struct.unpack('<H', dfusuffix[0:2])[0]
    #
    #     if missing_suffix:
    #         if check_suffix == SuffixReq.NEEDS_SUFFIX:
    #             _logger.info(f"{reason}")
    #             _logger.info("A valid DFU suffix will be required in a future dfu-util release")
    #             # raise ValueError("Valid DFU suffix needed")
    #
    #     elif check_suffix == SuffixReq.MAYBE_SUFFIX:
    #         _logger.warning(f"{reason}")


def store_file(file: DFUFile011, write_suffix: int, write_prefix: int):
    crc = 0xffffffff

    try:
        with open(file.name, 'wb') as file.file_p:

            # write prefix, if any
            if write_prefix:
                if file.prefix_type == PrefixType.LMDFU_PREFIX:
                    lmdfu_prefix = bytearray(LMDFU_PREFIX_LENGTH)
                    addr = file.lmdfu_address // 1024

                    # lmdfu_dfu_prefix payload length excludes prefix and suffix
                    len_ = file.size.total - file.size.prefix - file.size.suffix

                    lmdfu_prefix[0] = 0x01  # STELLARIS_DFU_PROG
                    lmdfu_prefix[1] = 0x00  # Reserved
                    lmdfu_prefix[2] = addr & 0xff
                    lmdfu_prefix[3] = addr >> 8
                    lmdfu_prefix[4] = len_ & 0xff
                    lmdfu_prefix[5] = (len_ >> 8) & 0xff
                    lmdfu_prefix[6] = (len_ >> 16) & 0xff
                    lmdfu_prefix[7] = (len_ >> 24)

                    crc = write_crc(file.file_p, crc, lmdfu_prefix, LMDFU_PREFIX_LENGTH)

                if file.prefix_type == PrefixType.LPCDFU_UNENCRYPTED_PREFIX:
                    lpcdfu_prefix = bytearray(LPCDFU_PREFIX_LENGTH)

                    # Payload is firmware and prefix rounded to 512 bytes
                    len_ = (file.size.total - file.size.suffix + 511) // 512

                    lpcdfu_prefix[0] = 0x1a  # Unencypted
                    lpcdfu_prefix[1] = 0x3f  # Reserved
                    lpcdfu_prefix[2] = (len_ & 0xff)
                    lpcdfu_prefix[3] = (len_ >> 8) & 0xff
                    for i in range(12, LPCDFU_PREFIX_LENGTH):
                        lpcdfu_prefix[i] = 0xff

                    crc = write_crc(file.file_p, crc, lpcdfu_prefix, LPCDFU_PREFIX_LENGTH)

            # write firmware binary
            crc = write_crc(file.file_p, crc, file.firmware[file.size.prefix:],
                            file.size.total + file.size.prefix + file.size.suffix)

            # write suffix, if any
            if write_suffix:
                dfusuffix = bytearray(DFU_SUFFIX_LENGTH)

                dfusuffix[0] = file.bcdDevice & 0xff
                dfusuffix[1] = file.bcdDevice >> 8
                dfusuffix[2] = file.idProduct & 0xff
                dfusuffix[3] = file.idProduct >> 8
                dfusuffix[4] = file.idVendor & 0xff
                dfusuffix[5] = file.idVendor >> 8
                dfusuffix[6] = file.bcdDFU & 0xff
                dfusuffix[7] = file.bcdDFU >> 8
                dfusuffix[8] = ord('U')
                dfusuffix[9] = ord('F')
                dfusuffix[10] = ord('D')
                dfusuffix[11] = DFU_SUFFIX_LENGTH

                crc = write_crc(file.file_p, crc, dfusuffix, DFU_SUFFIX_LENGTH - 4)

                dfusuffix[12] = crc
                dfusuffix[13] = crc >> 8
                dfusuffix[14] = crc >> 16
                dfusuffix[15] = crc >> 24

                crc = write_crc(file.file_p, crc, dfusuffix[12:], 4)

    except OSError as err:
        _logger.debug(err)
        raise OSError(f"Could not open file {file.name} for writing")


def show_suffix_and_prefix(file: DFUFile011) -> None:

    if file.size.prefix == LPCDFU_PREFIX_LENGTH:
        print(f"The file {file.name} contains a TI Stellaris "
              f"DFU prefix with the following properties:")
        print(f"Address:\t0x{file.lmdfu_address:08x}")
    elif file.size.prefix == LPCDFU_PREFIX_LENGTH:
        prefix = file.firmware
        size_kib = prefix[2] >> 1 | prefix[3] << 7
        print(f"The file {file.name} contains a NXP unencrypted "
              f"LPC DFU prefix with the following properties:")
        print(f"Size:\t{size_kib:5} kiB")
    elif file.size.prefix != 0:
        print(f"The file {file.name} contains an unknown prefix")

    if file.size.suffix > 0:
        print(f"The file {file.name} contains a DFU suffix with the following properties:")
        print(f"BCD device:\t0x{file.bcdDevice:04X}")
        print(f"Product ID:\t0x{file.idProduct:04X}")
        print(f"Vendor ID:\t0x{file.idVendor:04X}")
        print(f"BCD DFU:\t0x{file.bcdDFU:04X}")
        print(f"Length:\t\t{file.size.suffix}")
        print(f"CRC:\t\t0x{file.dwCRC:08X}")


def parse_dfu_suffix(file: DFUFile) -> int:
    """
    reads the file_p and name member, fills in all others
    FIXME: deprecated
    :param file:
    :return: 0 if no DFU suffix, positive if valid DFU suffix, negative on file read error
    """
    warnings.warn("parse_dfu_suffix is deprecated", FutureWarning)

    crc = 0xffffffff
    dfu_suffix = bytearray([0] * DFU_SUFFIX_LENGTH)

    try:
        with io.FileIO(file.name, 'rb') as file.file_p:
            file.size = file.file_p.seek(0, os.SEEK_END)
            file.file_p.seek(0)

            if file.size < DFU_SUFFIX_LENGTH:
                _logger.error("File too short for DFU suffix")
                return 0

            firmware = bytearray(file.file_p.read(file.size))

            for i in range(file.size - 4):
                crc = crc32_byte(crc, firmware[i])

            del firmware

            file.file_p.seek(-DFU_SUFFIX_LENGTH, os.SEEK_END)
            ret = file.file_p.readinto(dfu_suffix)

            if ret < 0:
                _logger.error("Could not read DFU suffix")
                return ret
            if ret < DFU_SUFFIX_LENGTH:
                _logger.error("Could not read whole DFU suffix")
                return -1

            suffix = _suffix.parse(dfu_suffix)
            file.dwCRC = suffix.dwCRC

            if file.dwCRC != crc:
                _logger.error("DFU CRC does not match")
                return 0

            file.bcdDFU = suffix.bcdDFU
            _logger.info(f"Dfu suffix version {hex(file.bcdDFU)}")

            file.suffix_len = dfu_suffix[11]
            if file.suffix_len < DFU_SUFFIX_LENGTH:
                _logger.error(f"Unsupported DFU suffix length {file.suffix_len}")
                return 0

            file.idVendor = suffix.idVendor
            file.idProduct = suffix.idProduct
            file.bcdDevice = suffix.bcdDevice

    except StreamError as e:
        _logger.error(f"Could not read whole DFU suffix, {e}")
        ret = -1
    except ConstError as e:
        _logger.error(f"No valid DFU suffix signature, {e}")
        ret = 0
    except Exception as e:
        _logger.exception(e)
        ret = -1
    return ret


def generate_dfu_suffix(file: DFUFile) -> int:
    """
    reads file, generates CRC and adds DFU suffix to file
    :param file:
    :return: positive on success, negative on errors
    """
    warnings.warn("generate_dfu_suffix is deprecated", FutureWarning)

    file.size = 0
    file.dwCRC = 0xffffffff
    file.suffix_len = DFU_SUFFIX_LENGTH
    file.bcdDFU = 0x0100  # Default to bcdDFU version 1.0

    suffix_data = {
        'bcdDevice': file.bcdDevice,
        'idProduct': file.idProduct,
        'idVendor': file.idVendor,
        'bcdDFU': file.bcdDFU,
        'bLength': file.suffix_len
    }

    dfu_suffix = bytearray(_suffix.build(suffix_data))

    try:
        with io.FileIO(file.name, 'rb+') as file_p:
            file.size = file_p.seek(0, os.SEEK_END)
            file_p.seek(0)

            # Make space for all but CRC
            firmware = bytearray(file_p.read(file.size))
            firmware.extend(dfu_suffix[:12])
            # Calculate CRC. It is calculated over file and suffix excluding the CRC itself
            for i in range(file.size + file.suffix_len - 4):
                file.dwCRC = crc32_byte(file.dwCRC, firmware[i])

            del firmware

            dfu_suffix[12:16] = ByteSwapped(_suffix.subcon.dwCRC).build(file.dwCRC)

            # Move to the end of the file
            file_p.seek(0, os.SEEK_END)

            # Add the suffix at the end of the file
            ret = file_p.write(dfu_suffix)

    except Exception as e:
        _logger.exception(e)
        ret = -1

    return ret
